# Author: Hubert Kario, (c) 2015
# Released under Gnu GPL v2.0, see LICENSE file for details

"""Parsing and processing of received TLS messages"""
from __future__ import print_function

import itertools
from functools import partial
import sys
import time

import tlslite.utils.tlshashlib as hashlib
from tlslite.constants import ContentType, HandshakeType, CertificateType,\
        HashAlgorithm, SignatureAlgorithm, ExtensionType,\
        SSL2HandshakeType, CipherSuite, GroupName, AlertDescription, \
        SignatureScheme, TLS_1_3_HRR, HeartbeatMode, \
        TLS_1_1_DOWNGRADE_SENTINEL, TLS_1_2_DOWNGRADE_SENTINEL, \
        HeartbeatMessageType, ClientCertificateType, CertificateStatusType, \
        CertificateCompressionAlgorithm, \
        ECPointFormat
from tlslite.messages import ServerHello, Certificate, ServerHelloDone,\
        ChangeCipherSpec, Finished, Alert, CertificateRequest, ServerHello2,\
        ServerKeyExchange, ClientHello, ServerFinished, CertificateStatus, \
        CertificateVerify, EncryptedExtensions, NewSessionTicket, Heartbeat,\
        KeyUpdate, HelloRequest, NewSessionTicket1_0, CompressedCertificate
from tlslite.extensions import TLSExtension, ALPNExtension,\
        ECPointFormatsExtension
from tlslite.utils.codec import Parser, Writer
from tlslite.utils.compat import b2a_hex
from tlslite.utils.cryptomath import secureHMAC, derive_secret, \
        HKDF_expand_label
from tlslite.mathtls import RFC7919_GROUPS, FFDHE_PARAMETERS, calc_key
from tlslite.keyexchange import KeyExchange, DHE_RSAKeyExchange, \
        ECDHE_RSAKeyExchange, AECDHKeyExchange
from tlslite.x509 import X509
from tlslite.x509certchain import X509CertChain
from tlslite.errors import TLSDecryptionFailed, TLSIllegalParameterException
from tlslite.handshakehashes import HandshakeHashes
from tlslite.handshakehelpers import HandshakeHelpers
from .handshake_helpers import calc_pending_states, kex_for_group, \
        curve_name_to_hash_tls13
from .helpers import ECDSA_SIG_TLS1_3_ALL
from .tree import TreeNode

# pylint: disable=import-error,no-name-in-module
# pylint: disable=bad-option-value,deprecated-class
if sys.version_info >= (3, 3):
    from collections.abc import Iterable
else:
    from collections import Iterable
# pylint: enable=bad-option-value,deprecated-class
# pylint: enable=import-error,no-name-in-module


class Expect(TreeNode):
    """Base class for objects handling message readers"""

    def __init__(self, content_type):
        """Prepare the class for handling tree graph"""
        super(Expect, self).__init__()
        self.content_type = content_type

    def is_expect(self):
        """Flag to tell if the object is a message processor"""
        return True

    def is_command(self):
        """Flag to tell that the object is a message processor"""
        return False

    def is_generator(self):
        """Flag to tell that the object is not a message generator"""
        return False

    def is_match(self, msg):
        """
        Checks if the object can handle message

        Note that the msg is a raw, unparsed message of indicated type that
        requires calling write() to get a raw bytearray() representation of it

        :type msg: tlslite.messages.Message
        :param msg: raw message to check
        """
        if msg.contentType == self.content_type:
            return True

        return False

    def process(self, state, msg):
        """
        Process the message and update the state accordingly.

        :type state: tlsfuzzer.runner.ConnectionState
        :param state: current connection state, needs to be updated after
            parsing the message by inheriting classes
        :type msg: tlslite.messages.Message
        :param msg: raw message to parse
        """
        raise NotImplementedError("Subclasses need to implement this!")


class ExpectMessage(Expect):
    """Common methods for handling TLS messages."""

    @staticmethod
    def _cmp_eq(our, recv, field_type=None, f_str=None):
        """
        Check if expected value matched received, if defined.

        If our is not None, compare with recv. If they don't match, try
        translating them with field_type.toStr() method and rise
        AssertionError with message formatted with f_str. First parameter
        to .format() will be expected value and the second one will be the
        received one
        """
        if our is None or our == recv:
            return

        if field_type:
            expected = field_type.toStr(our)
            received = field_type.toStr(recv)
        else:
            expected = our
            received = recv

        if not f_str:
            f_str = "Expected: {0}, received: {1}"
        raise AssertionError(f_str.format(expected, received))

    @classmethod
    def _cmp_eq_or_in(cls, our, recv, field_type=None, f_str=None):
        """
        Check if received value equals expected or is in expected list.

        If our is a list or set, check if recv is in it.
        If our is not None, check if it's equal to recv.
        If they don't match or are not part of a set, try translating
        them with field_type.toStr() method and raise AssertionError
        formatted with f_str. First parameter to .format() will be
        the expected value and the second one witll be the
        received one.
        """
        if our is None:
            return
        try:
            if recv in our:
                return
        except TypeError:
            return cls._cmp_eq(our, recv, field_type, f_str)

        # doesn't match, so prepare the error message
        if field_type:
            expected = "({0})".format(", ".join(
                field_type.toStr(i) for i in our))
            received = field_type.toStr(recv)
        else:
            expected = our
            received = recv

        if not f_str:
            f_str = "Received value ({1}) not in expected list: {0}"
        raise AssertionError(f_str.format(expected, received))

    @staticmethod
    def _cmp_eq_list(our, recv, field_type=None, f_str=None):
        """
        Check if expected list of values matched received, if defined.

        If our is not None, compare with recv. If they don't match, try
        translating items in the lists with field_type.toStr() method and rise
        AssertionError with message formatted with f_str. First parameter
        to .format() will be list of expected values and the second one will be
        the received one
        """
        if our is None or our == recv:
            return

        if field_type:
            expected = ", ".join(field_type.toStr(i) for i in our)
            expected = "({0})".format(expected)
            received = ", ".join(field_type.toStr(i) for i in recv)
            received = "({0})".format(received)
        else:
            expected = repr(our)
            received = repr(recv)

        if not f_str:
            f_str = "Expected: {0}, received: {1}"
        raise AssertionError(f_str.format(expected, received))


class ExpectHandshake(ExpectMessage):
    """Common methods for handling TLS Handshake protocol messages"""

    def __init__(self, content_type, handshake_type):
        """
        Set the type of message

        :type content_type: int
        :type handshake_type: int
        """
        super(ExpectHandshake, self).__init__(content_type)
        self.handshake_type = handshake_type

    def is_match(self, msg):
        """Check if message is a given type of handshake protocol message"""
        if not super(ExpectHandshake, self).is_match(msg):
            return False

        if not msg.write():  # if message is empty
            return False

        hs_type = Parser(msg.write()).get(1)
        if hs_type != self.handshake_type:
            return False

        return True

    def process(self, state, msg):
        raise NotImplementedError("Subclass need to implement this!")


def srv_ext_handler_ems(state, extension):
    """Process Extended Master Secret extension from server."""
    if extension.extData:
        raise AssertionError("Malformed EMS extension, data in payload")

    state.extended_master_secret = True


def srv_ext_handler_etm(state, extension):
    """Process Encrypt then MAC extension from server."""
    if extension.extData:
        raise AssertionError("Malformed EtM extension, data in payload")

    state.encrypt_then_mac = True


def srv_ext_handler_sni(state, extension):
    """Process the server_name extension from server."""
    del state  # kept for comatibility
    if extension.extData:
        raise AssertionError("Malformed SNI extenion, data in payload")


def srv_ext_handler_renego(state, extension):
    """Process the renegotiation_info from server."""
    if extension.renegotiated_connection != \
            state.key['client_verify_data'] + state.key['server_verify_data']:
        raise AssertionError("Invalid data in renegotiation_info")


def srv_ext_handler_alpn(state, extension):
    """Process the ALPN extension from server."""
    cln_hello = state.get_last_message_of_type(ClientHello)
    cln_ext = cln_hello.getExtension(ExtensionType.alpn)
    # the sent extension might have been provided with explicit encoding
    cln_ext = ALPNExtension().parse(Parser(cln_ext.extData))

    if not extension.protocol_names or len(extension.protocol_names) != 1:
        raise AssertionError("Malformed ALPN extension")
    if extension.protocol_names[0] not in cln_ext.protocol_names:
        raise AssertionError("Server selected ALPN protocol we did not "
                             "advertise")


def srv_ext_handler_ec_point(state, extension):
    """Process the ec_point_formats extension from server."""
    del state
    if extension.formats is None or not extension.formats:
        raise AssertionError("Malformed ec_point_formats extension")


def srv_ext_handler_npn(state, extension):
    """Process the NPN extension from server."""
    del state
    if extension.protocols is None or not extension.protocols:
        raise AssertionError("Malformed NPN extension")


def srv_ext_handler_session_ticket(state, extension):
    """Process the session_ticket extension from server."""
    del state
    if extension.ticket != b"":
        raise AssertionError("Malformed session_ticket extension")


def srv_ext_handler_key_share(state, extension):
    """Process the key_share extension from server."""
    cln_hello = state.get_last_message_of_type(ClientHello)
    cln_ext = cln_hello.getExtension(ExtensionType.key_share)

    group_id = extension.server_share.group

    cl_ext = next((i for i in cln_ext.client_shares if i.group == group_id),
                  None)
    if cl_ext is None:
        raise AssertionError("Server selected group we didn't advertise: {0}"
                             .format(GroupName.toStr(group_id)))

    kex = kex_for_group(group_id, state.version)

    state.key['ServerHello.extensions.key_share.key_exchange'] = \
        extension.server_share.key_exchange

    if not cl_ext.private:
        raise ValueError("private value for key share of group {0} missing"
                         .format(GroupName.toStr(group_id)))
    z = kex.calc_shared_key(cl_ext.private,
                            extension.server_share.key_exchange)

    state.key['DH shared secret'] = z


def hrr_ext_handler_key_share(state, extension):
    """Process the key_share extension in HRR message."""
    cln_hello = state.get_last_message_of_type(ClientHello)
    cln_ext = cln_hello.getExtension(ExtensionType.supported_groups)

    group_id = extension.selected_group

    if group_id not in cln_ext.groups:
        raise AssertionError("Server selected group we didn't advertise: {0}"
                             .format(GroupName.toStr(group_id)))


def hrr_ext_handler_cookie(state, extension):
    """Process the cookie extension in HRR message."""
    del state
    if not extension.cookie:
        raise AssertionError("Server sent empty cookie extension")


def srv_ext_handler_supp_vers(state, extension):
    """Process the supported_versions from server."""
    cln_hello = state.get_last_message_of_type(ClientHello)
    cln_ext = cln_hello.getExtension(ExtensionType.supported_versions)

    vers = extension.version

    if vers not in cln_ext.versions:
        raise AssertionError("Server selected version we didn't advertise: {0}"
                             .format(vers))

    state.version = vers


def srv_ext_handler_supp_groups(state, extension):
    """Process the supported_groups from server."""
    del state
    if not extension.groups:
        raise AssertionError("Server did not send any supported_groups")


def srv_ext_handler_status_request(state, extension):
    """
    Process the status_request extension from server.

    TLS 1.2 ServerHello specific, in TLS 1.3 the extension resides in
    Certificate message.
    """
    del state
    if extension.status_type is not None or \
            extension.responder_id_list != [] or \
            extension.request_extensions != bytearray():
        raise AssertionError("Server did send non empty status_request "
                             "extension")


def srv_ext_handler_heartbeat(state, extension):
    """Process the heartbeat extension from server."""
    del state
    if not extension.mode:
        raise AssertionError("Empty mode in heartbeat extension.")
    if extension.mode != HeartbeatMode.PEER_ALLOWED_TO_SEND and \
       extension.mode != HeartbeatMode.PEER_NOT_ALLOWED_TO_SEND:
        raise AssertionError("Invalid mode in heartbeat extension.")


def _srv_ext_handler_psk(state, extension, psk_configs):
    """Process the pre_shared_key extension from server.

    Since it needs the psk_configurations, it can't do it automatically
    so it shouldn't be part of _srv_ext_handler.
    """
    cln_hello = state.get_last_message_of_type(ClientHello)
    cln_ext = cln_hello.getExtension(ExtensionType.pre_shared_key)

    # the selection is 0-based
    if extension.selected >= len(cln_ext.identities):
        raise AssertionError("Server selected PSK we didn't send")

    ident = cln_ext.identities[extension.selected].identity
    if state.session_tickets:
        nst = state.session_tickets[-1]
        if nst.ticket == ident:
            state.key['PSK secret'] = HandshakeHelpers.calc_res_binder_psk(
                cln_ext.identities[extension.selected],
                state.key['resumption master secret'],
                [nst])
            return
    secret = next((i[1] for i in psk_configs if i[0] == ident), None)
    if not secret:
        raise ValueError("psk_configs are missing identity")

    state.key['PSK secret'] = secret


def gen_srv_ext_handler_psk(psk_configs=tuple()):
    """Creates a handler for pre_shared_key extension from the server."""
    return partial(_srv_ext_handler_psk, psk_configs=psk_configs)


def _srv_ext_handler_record_limit(state, extension, size=None):
    """Process record_size_limit extension from server."""
    cln_hello = state.get_last_message_of_type(ClientHello)
    cln_ext = cln_hello.getExtension(ExtensionType.record_size_limit)

    assert extension.record_size_limit is not None
    assert 64 <= extension.record_size_limit <= 2**14 + \
        int(state.version > (3, 3))

    if size and extension.record_size_limit != size:
        raise AssertionError("Server sent unexpected size in extension, "
                             "expected size: {0}, received size: {1}"
                             .format(size, extension.record_size_limit))

    if state.version <= (3, 3):
        # in TLS 1.2 and earlier we need to delay that to processing of
        # server CCS
        state._peer_record_size_limit = extension.record_size_limit
        state._our_record_size_limit = min(2**14, cln_ext.record_size_limit)
    else:
        # in TLS 1.3 we need to implement it right away (as the extension
        # applies only to encrypted messages)
        # the RecordLayer expects value that excludes content type
        state.msg_sock.recv_record_limit = min(
            2**14,
            cln_ext.record_size_limit-1)
        # this is just hint for padding callback
        state.msg_sock.send_record_limit = min(
            2**14,
            extension.record_size_limit-1)
        # this guides fragmentation
        state.msg_sock.recordSize = state.msg_sock.send_record_limit


def gen_srv_ext_handler_record_limit(size=None):
    """
    Create a handler for record_size_limit_extension from the server.

    Note that if the extension is actually negotiated, it will override
    any `~SetMaxRecordSize()` before EncryptedExtensions in TLS 1.3 and
    before ChangeCipherSpec in TLS 1.2 and earlier.

    :param int size: expected value from server, None for any valid
    """
    return partial(_srv_ext_handler_record_limit, size=size)


def clnt_ext_handler_status_request(state, extension):
    """
    Check status_request extension from initiating side.

    To be used in CertificateRequest
    """
    del state  # kept for compatibility
    if extension.extData:
        raise AssertionError(
            "Unexpected payload in status_request extension: {0}"
            .format(extension.extData))


def clnt_ext_handler_sig_algs(state, extension):
    """
    Check signature_algorithms or signature_algorithms_cert extension.

    To be used in ClientHello and CertificateRequest.
    """
    del state  # kept for API compatibility
    if not extension.sigalgs:
        raise AssertionError(
            "Empty or malformed {0} extension"
            .format(ExtensionType.toStr(extension.extType)))


_srv_ext_handler = \
        {ExtensionType.extended_master_secret: srv_ext_handler_ems,
         ExtensionType.encrypt_then_mac: srv_ext_handler_etm,
         ExtensionType.server_name: srv_ext_handler_sni,
         ExtensionType.renegotiation_info: srv_ext_handler_renego,
         ExtensionType.alpn: srv_ext_handler_alpn,
         ExtensionType.session_ticket: srv_ext_handler_session_ticket,
         ExtensionType.ec_point_formats: srv_ext_handler_ec_point,
         ExtensionType.supports_npn: srv_ext_handler_npn,
         ExtensionType.key_share: srv_ext_handler_key_share,
         ExtensionType.supported_versions: srv_ext_handler_supp_vers,
         ExtensionType.heartbeat: srv_ext_handler_heartbeat,
         ExtensionType.record_size_limit: _srv_ext_handler_record_limit,
         ExtensionType.status_request: srv_ext_handler_status_request}


_HRR_EXT_HANDLER = \
        {ExtensionType.key_share: hrr_ext_handler_key_share,
         ExtensionType.cookie: hrr_ext_handler_cookie}


_EE_EXT_HANDLER = \
        {ExtensionType.server_name: srv_ext_handler_sni,
         ExtensionType.alpn: srv_ext_handler_alpn,
         ExtensionType.supported_groups: srv_ext_handler_supp_groups,
         ExtensionType.heartbeat: srv_ext_handler_heartbeat,
         ExtensionType.record_size_limit: _srv_ext_handler_record_limit}


_CR_EXT_HANDLER = \
        {ExtensionType.status_request: clnt_ext_handler_status_request,
         ExtensionType.signature_algorithms: clnt_ext_handler_sig_algs,
         ExtensionType.signature_algorithms_cert: clnt_ext_handler_sig_algs}


class _ExpectExtensionsMessage(ExpectHandshake):
    """
    Common methods of messages that have a list of extensions.

    Used in ServerHello, EncryptedExtensions and CertificateRequest (in
    TLS 1.3)
    """
    def __init__(self, content_type, msg_type, extensions):
        super(_ExpectExtensionsMessage, self).__init__(
            content_type, msg_type)
        self.extensions = extensions

    def _compare_extensions(self, message):
        """
        Verify that server provided extensions match exactly expected list.
        """
        # if the list of extensions is present, make sure it matches exactly
        # with what the server sent
        if self.extensions and not message.extensions:
            raise AssertionError("Server did not send any extensions")
        if self.extensions is not None and message.extensions:
            expected = set(self.extensions.keys())
            got = set(i.extType for i in message.extensions)
            if got != expected:
                diff = expected.difference(got)
                if diff:
                    raise AssertionError("Server did not send extension(s): "
                                         "{0}".format(
                                             ", ".join((ExtensionType.toStr(i)
                                                        for i in diff))))
                diff = got.difference(expected)
                # we already checked if got != expected so diff here
                # must be non-empty if the one checked above is
                assert diff
                raise AssertionError("Server sent unexpected extension(s):"
                                     " {0}".format(
                                         ", ".join(ExtensionType.toStr(i)
                                                   for i in diff)))


class ExpectServerHello(_ExpectExtensionsMessage):
    """
    Parsing TLS Handshake protocol Server Hello messages.

    Processing of the ServerHello message updates the record layer
    to the version advertisied by the server.
    Use :py:class:`~tlsfuzzer.messages.SetRecordVersion` to change it earlier
    to send records with different versions.

    .. note::
      Receiving of the ServerHello in TLS 1.3 influences record layer
      encryption. After the message is received, the
      ``client_handshake_traffic_secret`` and
      ``server_handshake_traffic_secret``
      is derived and record layer is configured to expect encrypted records
      on the *receiving* side.

    :ivar str ~.description: identifier to print when processing of the
        node fails
    """

    def __init__(self, extensions=None, version=None, resume=False,
                 cipher=None, server_max_protocol=None, force_resume=False,
                 description=None):
        """
        Initialize the object

        :param dict extensions: extension objects to match the server sent
        extensions or callbacks to process and verify them. None means use
        automatic handlers that will verify the response against the extensions
        sent in ClientHello. Empty dict means that the server is expected to
        send no extensions. Order does not matter, but all extensions present
        and only extensions present in the list must be sent by server. None
        as the value of the relevant extension type can be used to select
        autohandler for a given extension type.

        :param tuple version: the literal version in the Server Hello message
        (needs to be (3, 3) for TLS 1.3, use extensions to expect TLS 1.3
        negotiation)

        :param tuple server_max_protocol: the higher protocol version supported
        by server. Used for testing downgrade signaling of servers.

        :type cipher: int or set-like
        :param int cipher: the id of the cipher that is expected to be
        negotiated by server. Can also be a list or set (needs to support
        ``in``) for a set of allowed ciphers.
        None (the default) means any valid cipher
        (i.e. not SCSV or GREASE) sent in ClientHello can be selected by
        server.

        :type resume: boolean
        :param resume: whether the session id should match the one from
        current state - IOW, if the server hello should belong to a resumed
        session. TLS 1.2 and earlier only. In TLS 1.3 resumption is handled
        by providing handler for ``pre_shared_key`` extension.

        :param boolean force_resume: assume that the session is getting resumed,
            even if the sessionID is empty. Applicable to TLS 1.2 and earlier
            only when using session tickets and not sending a sessionID.
        """
        super(ExpectServerHello, self).__init__(ContentType.handshake,
                                                HandshakeType.server_hello,
                                                extensions)
        self.cipher = cipher
        self.version = version
        self.resume = resume
        self.srv_max_prot = server_max_protocol
        self.force_resume = force_resume
        self.description = description

    def __str__(self):
        """Return human redable representation of the object."""
        if self.description:
            return "ExpectServerHello(description={0!r})"\
                   .format(self.description)
        return "ExpectServerHello()"

    @staticmethod
    def _get_autohandler(ext_id):
        try:
            return _srv_ext_handler[ext_id]
        except KeyError:
            raise AssertionError("No autohandler for "
                                 "{0}"
                                 .format(ExtensionType
                                         .toStr(ext_id)))

    def _process_extensions(self, state, cln_hello, srv_hello):
        """Check if extensions are correct."""
        # extensions allowed in TLS 1.3 ServerHello and HelloRetryRequest
        # messages (as some need to be echoed by server in EncryptedExtensions
        # and some in Certificate)
        sh_supported = [ExtensionType.pre_shared_key,
                        ExtensionType.supported_versions,
                        ExtensionType.key_share]
        hrr_supported = [ExtensionType.cookie,
                         ExtensionType.supported_versions,
                         ExtensionType.key_share]
        for ext in srv_hello.extensions:
            ext_id = ext.extType
            if state.version > (3, 3) and \
                    ((srv_hello.random != TLS_1_3_HRR and
                      ext_id not in sh_supported) or
                     (srv_hello.random == TLS_1_3_HRR and
                      ext_id not in hrr_supported)):
                raise AssertionError("Server sent unallowed "
                                     "extension of type {0}"
                                     .format(ExtensionType
                                             .toStr(ext_id)))
            # in TLS 1.2 generally the server can reply to any client sent
            # extension, and all of them end in ClientHello
            cl_ext = cln_hello.getExtension(ext_id)
            if ext_id == ExtensionType.renegotiation_info and \
                    CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV \
                    in cln_hello.cipher_suites:
                cl_ext = True
            if isinstance(self, ExpectHelloRetryRequest) and \
                    ext_id == ExtensionType.cookie:
                cl_ext = True
            if cl_ext is None:
                raise AssertionError("Server sent unadvertised "
                                     "extension of type {0}"
                                     .format(ExtensionType
                                             .toStr(ext_id)))
            handler = None
            if self.extensions:
                handler = self.extensions[ext_id]

            # use automatic handlers for some extensions
            if handler is None:
                handler = self._get_autohandler(ext_id)

            if callable(handler):
                handler(state, ext)
            elif isinstance(handler, TLSExtension):
                if not handler == ext:
                    raise AssertionError("Expected extension not "
                                         "matched for type {0}, "
                                         "received: {1}"
                                         .format(ExtensionType
                                                 .toStr(ext_id),
                                                 ext))
            else:
                raise ValueError("Bad extension handler for id {0}"
                                 .format(ExtensionType.toStr(ext_id)))

    @staticmethod
    def _extract_version(msg):
        """Extract the real version from the message if TLS 1.3 is in use."""
        ext = msg.getExtension(ExtensionType.supported_versions)

        # RFC 8446 "legacy_version field MUST be set to 0x0303"
        if msg.server_version > (3, 3):
            raise ValueError("Server sent invalid version in legacy_version "
                             "field")

        if ext and msg.server_version == (3, 3):
            return ext.version

        return msg.server_version

    def process(self, state, msg):
        """
        Process the message and update state accordingly

        :type state: ConnectionState
        :param state: overall state of TLS connection

        :type msg: Message
        :param msg: TLS Message read from socket
        """
        assert msg.contentType == ContentType.handshake

        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.server_hello

        srv_hello = ServerHello()
        srv_hello.parse(parser)

        # extract important info
        state.server_random = srv_hello.random

        cln_hello = state.get_last_message_of_type(ClientHello)

        # check for session_id based session resumption
        if self.resume:
            assert state.session_id == srv_hello.session_id
        if self.force_resume or ((state.session_id == srv_hello.session_id
                or cln_hello.session_id == srv_hello.session_id) and
                srv_hello.session_id != bytearray(0) and
                self._extract_version(srv_hello) < (3, 4)):
            # TLS 1.2 resumption, TLS 1.3 is based on PSKs
            state.resuming = True
            assert state.cipher == srv_hello.cipher_suite
            assert state.version == self._extract_version(srv_hello)
        state.session_id = srv_hello.session_id

        self._cmp_eq(self.version, srv_hello.server_version,
                     f_str="Server selected unexpected protocol version. "
                           "Expected: {0}, received: {1}.")

        self._cmp_eq_or_in(
            self.cipher, srv_hello.cipher_suite,
            f_str="Server selected unexpected ciphersuite. "
                  "Expected: {0}, received: {1}.")

        # check if server sent cipher matches what we advertised in CH
        if srv_hello.cipher_suite not in cln_hello.cipher_suites:
            cipher = srv_hello.cipher_suite
            if cipher in CipherSuite.ietfNames:
                name = "{0} ({1:#06x})".format(CipherSuite.ietfNames[cipher],
                                               cipher)
            else:
                name = "{0:#06x}".format(cipher)
            raise AssertionError("Server responded with cipher we did"
                                 " not advertise: {0}".format(name))

        state.cipher = srv_hello.cipher_suite
        state.version = self._extract_version(srv_hello)

        # update the state of connection
        state.msg_sock.version = state.version
        state.msg_sock.tls13record = state.version > (3, 3)

        self._check_against_hrr(state, srv_hello)

        state.handshake_messages.append(srv_hello)
        state.handshake_hashes.update(msg.write())

        # Reset value of the session-wide settings
        state.extended_master_secret = False
        state.encrypt_then_mac = False

        self._check_downgrade_protection(srv_hello)

        self._compare_extensions(srv_hello)

        if srv_hello.extensions:
            self._process_extensions(state, cln_hello, srv_hello)

        if state.version > (3, 3):
            self._setup_tls13_handshake_keys(state)
        return srv_hello

    @staticmethod
    def _check_against_hrr(state, srv_hello):
        if state.version < (3, 4):
            return

        hrr = state.get_last_message_of_type(ServerHello)
        if not hrr or hrr.random != TLS_1_3_HRR:
            # not an HRR, so HRR tests don't apply to it
            return

        if hrr.cipher_suite != srv_hello.cipher_suite:
            raise AssertionError("Server picked different cipher suite than "
                                 "it advertised in HelloRetryRequest")

        hrr_version = hrr.getExtension(ExtensionType.supported_versions)
        sh_version = srv_hello.getExtension(ExtensionType.supported_versions)

        if hrr_version.version != sh_version.version:
            raise AssertionError("Server picked different protocol version "
                                 "than it advertised in HelloRetryRequest")

    def _setup_tls13_handshake_keys(self, state):
        """Set up the encryption keys for the TLS 1.3 handshake."""
        del self
        prf_name = state.prf_name
        prf_size = state.prf_size

        # Derive PSK secret
        psk = state.key.setdefault('PSK secret', bytearray(prf_size))

        # Derive TLS 1.3 early secret
        secret = bytearray(prf_size)
        secret = secureHMAC(secret, psk, prf_name)
        state.key['early secret'] = secret

        # Derive TLS 1.3 handshake secret
        secret = derive_secret(secret, b'derived', None, prf_name)
        dh_secret = state.key.setdefault('DH shared secret',
                                         bytearray(prf_size))
        secret = secureHMAC(secret, dh_secret, prf_name)
        state.key['handshake secret'] = secret

        # Derive TLS 1.3 traffic secrets
        s_traffic_secret = derive_secret(secret, b's hs traffic',
                                         state.handshake_hashes,
                                         prf_name)
        state.key['server handshake traffic secret'] = s_traffic_secret
        c_traffic_secret = derive_secret(secret, b'c hs traffic',
                                         state.handshake_hashes,
                                         prf_name)
        state.key['client handshake traffic secret'] = c_traffic_secret

        state.msg_sock.calcTLS1_3PendingState(
            state.cipher, c_traffic_secret, s_traffic_secret, None)

        state.msg_sock.changeReadState()

    def _check_downgrade_protection(self, srv_hello):
        """
        Verify that server provided downgrade protection as specified in
        RFC 8446, Section 4.1.3
        """
        # even if we don't know which version server supports, some values
        # are obviously incorrect:
        if (self._extract_version(srv_hello) > (3, 3) and
                srv_hello.random[24:] == TLS_1_2_DOWNGRADE_SENTINEL) or \
                (self._extract_version(srv_hello) > (3, 2) and
                 srv_hello.random[24:] == TLS_1_1_DOWNGRADE_SENTINEL):
            raise AssertionError(
                "Server set downgrade protection sentinel but shouldn't "
                "have done that")
        # as we're doing both TLS 1.2 tests and TLS 1.3 tests with `scripts/`
        # we don't know when setting the sentinel is expected and when
        # it is not as the negotiation might have ended up with TLS 1.2
        # because that was the highest version we advertised
        if self.srv_max_prot is None:
            return

        downgrade_value = None
        if self.srv_max_prot > (3, 3) \
                and self._extract_version(srv_hello) == (3, 3):
            downgrade_value = TLS_1_2_DOWNGRADE_SENTINEL
        elif self.srv_max_prot > (3, 2) \
                and self._extract_version(srv_hello) < (3, 3):
            downgrade_value = TLS_1_1_DOWNGRADE_SENTINEL
        else:
            if srv_hello.random[24:] == TLS_1_1_DOWNGRADE_SENTINEL or \
                srv_hello.random[24:] == TLS_1_2_DOWNGRADE_SENTINEL:
                raise AssertionError(
                    "Server set downgrade protection sentinel but shouldn't "
                    "have done that")

        if downgrade_value is not None:
            if srv_hello.random[24:] != downgrade_value:
                raise AssertionError(
                    "Server failed to set downgrade protection sentinel in "
                    "ServerHello.random value")


class ExpectHelloRetryRequest(ExpectServerHello):
    """Processing of the TLS 1.3 HelloRetryRequest message."""

    def __init__(self, extensions=None, version=None, cipher=None):
        super(ExpectHelloRetryRequest, self).__init__(
            extensions, version, cipher)
        self._ch_hh = None
        self._msg = None

    def process(self, state, msg):
        self._ch_hh = state.handshake_hashes.copy()
        self._msg = msg
        hrr = super(ExpectHelloRetryRequest, self).process(state, msg)
        assert hrr.random == TLS_1_3_HRR

    @staticmethod
    def _get_autohandler(ext_id):
        try:
            return _HRR_EXT_HANDLER[ext_id]
        except KeyError:
            try:
                return _srv_ext_handler[ext_id]
            except KeyError:
                raise AssertionError("No autohandler for {0}".format(
                    ExtensionType.toStr(ext_id)))

    def _setup_tls13_handshake_keys(self, state):
        """Prepare handshake ciphers for the HRR handling"""
        prf_name = state.prf_name

        ch_hash = self._ch_hh.digest(prf_name)
        new_hh = HandshakeHashes()
        writer = Writer()
        writer.add(HandshakeType.message_hash, 1)
        writer.addVarSeq(ch_hash, 1, 3)
        new_hh.update(writer.bytes)

        new_hh.update(self._msg.write())

        state.handshake_hashes = new_hh


class ExpectServerHello2(ExpectHandshake):
    """Processing of SSLv2 Handshake Protocol SERVER-HELLO message"""

    def __init__(self, version=None):
        c_type = ContentType.handshake
        h_type = SSL2HandshakeType.server_hello
        super(ExpectServerHello2, self).__init__(c_type,
                                                 h_type)
        self.version = version

    def process(self, state, msg):
        """
        Process the message and update state accordingly

        :type state: `~ConnectionState`
        :param state: overall state of TLS connection

        :type msg: Message
        :param msg: TLS Message read from socket
        """
        # the value is faked for SSLv2 protocol, but let's just check sanity
        assert msg.contentType == ContentType.handshake

        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == SSL2HandshakeType.server_hello

        server_hello = ServerHello2().parse(parser)

        state.handshake_messages.append(server_hello)
        state.handshake_hashes.update(msg.write())

        self._cmp_eq(self.version, server_hello.server_version,
                     f_str="Server picked unexpected protocol version."
                           "Expected: {0}, received: {1}.")

        if server_hello.session_id_hit:
            state.resuming = True
        state.session_id = server_hello.session_id
        state.server_random = server_hello.session_id
        state.version = server_hello.server_version
        state.msg_sock.version = server_hello.server_version

        # fake a certificate message so finding the server public key works
        x509 = X509()
        x509.parseBinary(server_hello.certificate)
        cert_chain = X509CertChain([x509])
        certificate = Certificate(CertificateType.x509)
        certificate.create(cert_chain)
        state.handshake_messages.append(certificate)
        # fake message so don't update handshake hashes


class ExpectCertificate(ExpectHandshake):
    """Processing TLS Handshake protocol Certificate messages"""

    def __init__(self, cert_type=CertificateType.x509):
        super(ExpectCertificate, self).__init__(ContentType.handshake,
                                                HandshakeType.certificate)
        self.cert_type = cert_type
        self._old_cert = None
        self._old_cert_bytes = None

    def process(self, state, msg):
        """
        :type state: `~ConnectionState`
        """
        assert msg.contentType == ContentType.handshake

        msg_bytes = msg.write()
        if self._old_cert_bytes is not None and \
                msg_bytes == self._old_cert_bytes:
            cert = self._old_cert
        else:
            parser = Parser(msg_bytes)
            hs_type = parser.get(1)
            assert hs_type == HandshakeType.certificate

            cert = Certificate(self.cert_type, state.version)
            cert.parse(parser)
            self._old_cert_bytes = msg_bytes
            self._old_cert = cert

        state.handshake_messages.append(cert)
        state.handshake_hashes.update(msg_bytes)


class ExpectCompressedCertificate(ExpectHandshake):
    def __init__(self, cert_type=CertificateType.x509, compression_algo=None):
        super(ExpectCompressedCertificate, self).__init__(
            ContentType.handshake,
            HandshakeType.compressed_certificate
        )
        self.cert_type = cert_type
        self._old_cert = None
        self._old_cert_bytes = None
        self._compression_algo = compression_algo

    def process(self, state, msg):
        """
        :type state: `~ConnectionState`
        """
        assert msg.contentType == ContentType.handshake

        msg_bytes = msg.write()
        if self._old_cert_bytes is not None and \
                msg_bytes == self._old_cert_bytes:
            cert = self._old_cert
        else:
            parser = Parser(msg_bytes)
            hs_type = parser.get(1)
            assert hs_type == HandshakeType.compressed_certificate

            cert = CompressedCertificate(self.cert_type, state.version)
            cert.parse(parser)
            self._old_cert_bytes = msg_bytes
            self._old_cert = cert

            if self._compression_algo is not None:
                if cert.compression_algo != self._compression_algo:
                    raise AssertionError(
                        "Compression algorithms doesn't much. " +
                        "Expected: {0}, Got: {1}".format(
                            self._compression_algo, cert.compression_algo)
                    )
            else:
                ch = state.get_last_message_of_type(ClientHello)
                assert ch is not None
                ext = ch.getExtension(ExtensionType.compress_certificate)
                assert ext is not None
                assert cert.compression_algo in ext.algorithms

        state.handshake_messages.append(cert)
        state.handshake_hashes.update(msg_bytes)


class ExpectCertificateVerify(ExpectHandshake):
    """
    Processing TLS Handshake protocol Certificate Verify messages.
    :param tuple(int,int) version: Expected TLS version of the message. If not
    provided will be taken from the state.
    :param tuple(int,int) sig_alg: Expected value of the signature scheme
    created by the server. If not provided it will be compared with signature
    algorithm extension from client hello.
    :param str hash_file: The file where hashes of the signature context will
    be logged
    :param str sig_file: The file where the signatures themselves will be
    logged

    """
    def __init__(
        self, version=None, sig_alg=None, hash_file=None, sig_file=None
    ):
        super(ExpectCertificateVerify, self).__init__(
            ContentType.handshake,
            HandshakeType.certificate_verify)
        self.version = version
        self.sig_alg = sig_alg
        self.hash_file = hash_file
        self.sig_file = sig_file

    def process(self, state, msg):
        """
        :type state: `~ConnectionState`
        """
        assert msg.contentType == ContentType.handshake
        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.certificate_verify

        if self.version is None:
            self.version = state.version

        cert_v = CertificateVerify(self.version)
        cert_v.parse(parser)

        if self.sig_alg:
            assert self.sig_alg == cert_v.signatureAlgorithm
        else:
            c_hello = state.get_last_message_of_type(ClientHello)
            ext = c_hello.getExtension(ExtensionType.signature_algorithms)
            assert cert_v.signatureAlgorithm in ext.sigalgs
            key_type = state.get_server_public_key().key_type
            if key_type == "rsa-pss":
                # in TLS 1.3 only RSA-PSS signatures are allowed
                assert cert_v.signatureAlgorithm in (
                    SignatureScheme.rsa_pss_pss_sha256,
                    SignatureScheme.rsa_pss_pss_sha384,
                    SignatureScheme.rsa_pss_pss_sha512)
            elif key_type == "rsa":
                # in TLS 1.3 only RSA-PSS signatures are allowed
                assert cert_v.signatureAlgorithm in (
                    SignatureScheme.rsa_pss_rsae_sha256,
                    SignatureScheme.rsa_pss_rsae_sha384,
                    SignatureScheme.rsa_pss_rsae_sha512)
            elif key_type in ("Ed25519", "Ed448"):
                assert cert_v.signatureAlgorithm in (
                    SignatureScheme.ed25519,
                    SignatureScheme.ed448)
                if getattr(SignatureScheme, key_type.lower()) != \
                        cert_v.signatureAlgorithm:
                    raise AssertionError(
                        "Mismatched signature ({0}) for used key ({1})"
                        .format(
                            SignatureScheme.toStr(cert_v.signatureAlgorithm),
                            key_type))
            else:
                assert key_type == "ecdsa"
                curve_name = state.get_server_public_key().curve_name
                assert curve_name in (
                    "NIST256p", "NIST384p", "NIST521p",
                    "BRAINPOOLP256r1", "BRAINPOOLP384r1", "BRAINPOOLP512r1"
                )
                sigalg = cert_v.signatureAlgorithm
                assert sigalg in ECDSA_SIG_TLS1_3_ALL
                if "NIST" in curve_name:
                    hash_name = curve_name_to_hash_tls13(curve_name)
                    # in TLS 1.3 the hash is bound to key curve
                    if sigalg != (getattr(HashAlgorithm, hash_name),
                                  SignatureAlgorithm.ecdsa):
                        raise AssertionError(
                            "Invalid signature type for {1} key, "
                            "received: {0}"
                            .format(SignatureScheme.toStr(sigalg), curve_name))
                else:
                    # but for Brainpool it's simpler to just explicitly list
                    if curve_name == "BRAINPOOLP256r1":
                        exp_sig_alg = \
                            SignatureScheme.ecdsa_brainpoolP256r1tls13_sha256
                    elif curve_name == "BRAINPOOLP384r1":
                        exp_sig_alg = \
                            SignatureScheme.ecdsa_brainpoolP384r1tls13_sha384
                    else:
                        assert curve_name == "BRAINPOOLP512r1"
                        exp_sig_alg = \
                            SignatureScheme.ecdsa_brainpoolP512r1tls13_sha512
                    if sigalg != exp_sig_alg:
                        raise AssertionError(
                            "Invalid signature type for {1} key, "
                            "received: {0}"
                            .format(SignatureScheme.toStr(sigalg), curve_name))

        salg = cert_v.signatureAlgorithm

        if salg in (SignatureScheme.ed25519, SignatureScheme.ed448):
            hash_name = "intrinsic"
            padding = None
            salt_len = None
        elif salg[1] == SignatureAlgorithm.ecdsa:
            hash_name = HashAlgorithm.toStr(salg[0])
            padding = None
            salt_len = None
        elif salg in (SignatureScheme.ecdsa_brainpoolP256r1tls13_sha256,
                      SignatureScheme.ecdsa_brainpoolP384r1tls13_sha384,
                      SignatureScheme.ecdsa_brainpoolP512r1tls13_sha512):
            scheme = SignatureScheme.toRepr(salg)
            hash_name = SignatureScheme.getHash(scheme)
            padding = None
            salt_len = None
        else:
            scheme = SignatureScheme.toRepr(salg)
            hash_name = SignatureScheme.getHash(scheme)
            padding = SignatureScheme.getPadding(scheme)
            salt_len = getattr(hashlib, hash_name)().digest_size

        transcript_hash = state.handshake_hashes.digest(state.prf_name)
        sig_context = bytearray(b'\x20' * 64 +
                                b'TLS 1.3, server CertificateVerify' +
                                b'\x00') + transcript_hash

        if not state.get_server_public_key().hashAndVerify(
                cert_v.signature,
                sig_context,
                padding,
                hash_name,
                salt_len):
            raise AssertionError("Signature verification failed")

        if self.hash_file:
            data = getattr(hashlib, hash_name)(sig_context).digest()
            self.hash_file.write(data)

        if self.sig_file:
            self.sig_file.write(cert_v.signature)

        state.handshake_messages.append(cert_v)
        state.handshake_hashes.update(msg.write())


class ExpectServerKeyExchange(ExpectHandshake):
    """Processing TLS Handshake protocol Server Key Exchange message"""

    def __init__(self, version=None, cipher_suite=None, valid_sig_algs=None,
                 valid_groups=None, valid_params=None):
        """
        Expect ServerKeyExchange message from server.

        :param list(int) valid_groups: TLS group identifiers for groups that
            server can use. In case the groups include identifiers between 256
            and 512 (see RFC 7919), the node will also check that the server
            selected FFDH parameters match the parameters specified in the RFC.

        :param set(tuple(int,int)) valid_params: set of explicit expected
            parameters used by the server, the first element of the tuple
            is the expected generator and the second is the prime used for the
            DH calculation. Applicable only to ciphersuites that use FFDHE
            key exchange.
        """
        msg_type = HandshakeType.server_key_exchange
        super(ExpectServerKeyExchange, self).__init__(ContentType.handshake,
                                                      msg_type)
        self.version = version
        self.cipher_suite = cipher_suite
        self.valid_sig_algs = valid_sig_algs
        self.valid_groups = valid_groups
        self.valid_params = valid_params
        if self.valid_groups and self.valid_params:
            raise ValueError("valid_groups and valid_params are exclusive")

    def _checkParams(self, server_key_exchange):
        groups = []
        if self.valid_groups and any(i in range(256, 512)
                                     for i in self.valid_groups):
            groups = [RFC7919_GROUPS[i - 256] for i in self.valid_groups
                      if i in range(256, 512)]
        if self.valid_params:
            groups = self.valid_params
        server_params = (server_key_exchange.dh_g, server_key_exchange.dh_p)
        if groups and server_params not in groups:
            for name, params in FFDHE_PARAMETERS.items():
                if server_params == params:
                    raise AssertionError(
                        "DH parameters not from valid set, "
                        "received: {0}".format(name))
            raise AssertionError(
                "DH parameters not from valid set, "
                "received: g:{0}, p:{1}".format(
                    hex(server_params[0]),
                    hex(server_params[1])))

    def process(self, state, msg):
        """Process the Server Key Exchange message"""
        assert msg.contentType == ContentType.handshake
        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.server_key_exchange

        if self.version is None:
            self.version = state.version
        if self.cipher_suite is None:
            self.cipher_suite = state.cipher
        valid_sig_algs = self.valid_sig_algs
        valid_groups = self.valid_groups

        server_key_exchange = ServerKeyExchange(self.cipher_suite,
                                                self.version)
        server_key_exchange.parse(parser)

        client_random = state.client_random
        server_random = state.server_random
        public_key = state.get_server_public_key()
        server_hello = state.get_last_message_of_type(ServerHello)
        client_hello = state.get_last_message_of_type(ClientHello)
        if server_hello is None:
            server_hello = ServerHello
            server_hello.server_version = state.version
        if valid_sig_algs is None:
            # if the value was unset in script, get the advertised value from
            # Client Hello
            if client_hello is not None:
                sig_algs_ext = client_hello.getExtension(ExtensionType.
                                                         signature_algorithms)
                if sig_algs_ext is not None:
                    valid_sig_algs = sig_algs_ext.sigalgs
            if valid_sig_algs is None:
                # no advertised means support for sha1 only
                valid_sig_algs = [(HashAlgorithm.sha1, SignatureAlgorithm.rsa)]
                if self.cipher_suite in CipherSuite.ecdheEcdsaSuites:
                    valid_sig_algs = [(HashAlgorithm.sha1,
                                       SignatureAlgorithm.ecdsa)]
                if self.cipher_suite in CipherSuite.dheDsaSuites:
                    valid_sig_algs = [(HashAlgorithm.sha1,
                                       SignatureAlgorithm.dsa)]

        try:
            KeyExchange.verifyServerKeyExchange(server_key_exchange,
                                                public_key,
                                                client_random,
                                                server_random,
                                                valid_sig_algs)
        except TLSDecryptionFailed:
            # very rarely validation of signature fails, print it so that
            # we have a chance in debugging it
            print("Bad signature: {0}"
                  .format(b2a_hex(server_key_exchange.signature)),
                  file=sys.stderr)
            raise

        if self.cipher_suite in CipherSuite.dhAllSuites:
            self._checkParams(server_key_exchange)
            state.key_exchange = DHE_RSAKeyExchange(self.cipher_suite,
                                                    clientHello=None,
                                                    serverHello=server_hello,
                                                    privateKey=None)
            state.key['ServerKeyExchange.key_share'] = \
                server_key_exchange.dh_Ys
            state.key['ServerKeyExchange.dh_p'] = server_key_exchange.dh_p
        elif self.cipher_suite in CipherSuite.ecdhAllSuites:
            # extract valid groups from Client Hello
            if valid_groups is None:
                client_hello = state.get_last_message_of_type(ClientHello)
                if client_hello is not None:
                    groups_ext = client_hello.getExtension(ExtensionType.
                                                           supported_groups)
                    if groups_ext is not None:
                        valid_groups = groups_ext.groups
                if valid_groups is None:
                    # no advertised means support for all
                    valid_groups = GroupName.allEC
            state.key_exchange = \
                ECDHE_RSAKeyExchange(self.cipher_suite,
                                     clientHello=client_hello,
                                     serverHello=server_hello,
                                     privateKey=None,
                                     acceptedCurves=valid_groups)
            state.key['ServerKeyExchange.key_share'] = \
                server_key_exchange.ecdh_Ys
        else:
            raise AssertionError("Unsupported cipher selected")
        state.key['premaster_secret'] = state.key_exchange.\
            processServerKeyExchange(public_key,
                                     server_key_exchange)

        state.handshake_messages.append(server_key_exchange)
        state.handshake_hashes.update(msg.write())


# RFC8446 Section 4.2 says that implementation MUST reject extensions
# it recognises but which are not allowed in CertificateRequest
# check it against all defined in RFC8446
TLS_1_3_CR_FORBIDDEN = set((
    ExtensionType.server_name,
    1,  # ExtensionType.max_fragment_length
    ExtensionType.supported_groups,
    14,  # ExtensionType.use_srtp
    ExtensionType.heartbeat,
    ExtensionType.alpn,
    19,  # ExtensionType.client_certificate_type
    20,  # ExtensionType.server_certificate_type
    21,  # ExtensionType.padding,
    ExtensionType.key_share,
    ExtensionType.pre_shared_key,
    ExtensionType.psk_key_exchange_modes,
    ExtensionType.early_data,
    ExtensionType.cookie,
    ExtensionType.supported_versions,
    49  # ExtensionType.post_handshake_auth
    ))


class ExpectCertificateRequest(_ExpectExtensionsMessage):
    """Processing TLS Handshake protocol Certificate Request message."""

    def __init__(self, sig_algs=None, cert_types=None,
                 sanity_check_cert_types=True, extensions=None, context=None):
        """
        Set expected parameters for the CertificateRequest message.

        :param sig_algs: a list of signature algorithms that we are expecting
            from server. Needs to be in-order and complete. ``None`` to accept
            any list from server. Applicable to TLS 1.2 and later only.
            Do not use together with non-default ``extensions``.
        :param cert_types: a list of client certificate types that we are
            expecting from server. Needs to be in-order and complete.
            ``None`` to accept any list from server. Applicable to TLS 1.2 and
            earlier only.
        :param sanity_check_cert_types: set to ``False`` to disable
            verification checking if every signature algorithm has a
            corresponding client certificate type.
        :param extensions: dictionary with extensions that need to be included
            in the message. Set to ``None`` to accept any, set to empty dict to
            expect no extensions. This has to be an exact match of the
            extensions included in the message. Usable in TLS 1.3 only.
        """
        msg_type = HandshakeType.certificate_request
        super(ExpectCertificateRequest, self).__init__(ContentType.handshake,
                                                       msg_type,
                                                       extensions)
        self.sig_algs = sig_algs
        self.cert_types = cert_types
        self.context = context
        self.sanity_check_cert_types = sanity_check_cert_types
        if sig_algs is not None and extensions is not None:
            raise ValueError("Can't set sig_algs and extensions at the same "
                             "time")

    @staticmethod
    def _sanity_check_cert_types(cert_request):
        """Verify that the CertificateRequest is self-consistent."""
        for sig_alg in cert_request.supported_signature_algs:
            if sig_alg in (SignatureScheme.ecdsa_brainpoolP256r1tls13_sha256,
                           SignatureScheme.ecdsa_brainpoolP384r1tls13_sha384,
                           SignatureScheme.ecdsa_brainpoolP512r1tls13_sha512):
                raise AssertionError(
                    "TLS 1.3 specific signature scheme in an earlier protocol "
                    "version: {0}".format(sig_alg))

            if sig_alg[1] == SignatureAlgorithm.ecdsa or \
                sig_alg in (SignatureScheme.ed25519,
                            SignatureScheme.ed448):
                key_type = "ECDSA"
                cert_type = "ecdsa_sign"
            elif sig_alg[1] == SignatureAlgorithm.rsa:
                key_type = "RSA"
                cert_type = "rsa_sign"
            elif sig_alg[1] == SignatureAlgorithm.dsa:
                key_type = "DSA"
                cert_type = "dss_sign"
            else:
                sig_scheme = SignatureScheme.toRepr(sig_alg)
                key_type = SignatureScheme.getKeyType(sig_scheme)
                assert key_type == "rsa", \
                    "Unsupported signature algorithm: {0}".format(sig_alg)
                cert_type = "rsa_sign"

            if getattr(ClientCertificateType, cert_type) \
                    not in cert_request.certificate_types:
                raise AssertionError(
                    "CertificateRequest includes {1} signature algorithms "
                    "({0}) but does not include {2} client "
                    "certificate type".format(sig_alg, key_type, cert_type))

    @staticmethod
    def _get_autohandler(ext_id):
        try:
            return _CR_EXT_HANDLER[ext_id]
        except KeyError:
            # handle future/GREASE extensions
            return None

    def _process_extensions(self, state, msg):
        for ext in msg.extensions:
            ext_id = ext.extType
            handler = None
            if ext_id in TLS_1_3_CR_FORBIDDEN:
                raise AssertionError(
                    "Server sent extension that is explicitly forbidden in "
                    "CertificateRequest messages: {0}".format(
                        ExtensionType.toStr(ext_id)))
            if self.extensions:
                handler = self.extensions[ext_id]
            if handler is None:
                handler = self._get_autohandler(ext_id)

            if callable(handler):
                handler(state, ext)
            elif isinstance(handler, TLSExtension):
                if not handler == ext:
                    raise AssertionError(
                        "Expected extension not matched for type {0}, "
                        "received: {1}".format(ExtensionType.toStr(ext_id),
                                               ext))
            elif handler is None:
                # since server can send arbitrary extensions, we need to
                # be able to process them, so if the self.extensions is unset
                # we can just do nothing
                pass
            else:
                raise ValueError("Bad extension handler for id {0}".format(
                    ExtensionType.toStr(ext_id)))

    def process(self, state, msg):
        """
        Check received Certificate Request

        :type state: ConnectionState
        """
        assert msg.contentType == ContentType.handshake

        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.certificate_request

        cert_request = CertificateRequest(state.version)
        cert_request.parse(parser)

        self._cmp_eq_list(self.sig_algs, cert_request.supported_signature_algs,
                          SignatureScheme,
                          f_str="Unexpected signature algorithms. Got: {1}, "
                                "expected: {0}")

        self._cmp_eq_list(self.cert_types, cert_request.certificate_types,
                          ClientCertificateType,
                          f_str="Unexpected client certificate types. Got: "
                                "{1}, expected: {0}")

        if state.version == (3, 3) and self.sanity_check_cert_types:
            # only in TLS 1.2 do the sig algs coexist with cert types
            self._sanity_check_cert_types(cert_request)

        if state.version >= (3, 4):
            self._compare_extensions(cert_request)
            self._process_extensions(state, cert_request)
            if self.context is not None:
                self.context.append(cert_request)

        state.handshake_messages.append(cert_request)
        state.handshake_hashes.update(msg.write())


class ExpectServerHelloDone(ExpectHandshake):
    """Processing TLS Handshake protocol ServerHelloDone messages"""

    def __init__(self):
        super(ExpectServerHelloDone,
              self).__init__(ContentType.handshake,
                             HandshakeType.server_hello_done)

    def process(self, state, msg):
        """
        :type state: ConnectionState
        :type msg: Message
        """
        assert msg.contentType == ContentType.handshake

        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.server_hello_done

        srv_hello_done = ServerHelloDone()
        srv_hello_done.parse(parser)

        state.handshake_messages.append(srv_hello_done)
        state.handshake_hashes.update(msg.write())


class ExpectChangeCipherSpec(Expect):
    """
    Processing TLS Change Cipher Spec messages.

    .. note::
      In SSLv3 up to TLS 1.2, the message modifies the state of record layer
      to expect encrypted records *after* receiving this message.
      In case of renegotiation, record layer will expect records encrypted
      with the newly negotiated keys. In TLS 1.3 it has no effect on record
      layer encryption.
    """

    def __init__(self):
        super(ExpectChangeCipherSpec,
              self).__init__(ContentType.change_cipher_spec)

    def process(self, state, msg):
        """
        :type state: ConnectionState
        :type msg: Message
        """
        assert msg.contentType == ContentType.change_cipher_spec
        parser = Parser(msg.write())
        ccs = ChangeCipherSpec().parse(parser)

        assert ccs.type == 1

        if state.version < (3, 4):
            # in TLS 1.3 the CCS does not have any affect on encryption
            if state.resuming:
                state.msg_sock.encryptThenMAC = state.encrypt_then_mac
                calc_pending_states(state)

            state.msg_sock.changeReadState()

            if state._our_record_size_limit:
                state.msg_sock.recv_record_limit = state._our_record_size_limit


class ExpectVerify(ExpectHandshake):
    """Processing of SSLv2 SERVER-VERIFY message"""

    def __init__(self):
        super(ExpectVerify, self).__init__(ContentType.handshake,
                                           SSL2HandshakeType.server_verify)

    def process(self, state, msg):
        """Check if the VERIFY message has expected value"""
        assert msg.contentType == ContentType.handshake
        parser = Parser(msg.write())

        msg_type = parser.get(1)
        assert msg_type == SSL2HandshakeType.server_verify


class ExpectFinished(ExpectHandshake):
    """
    Processing TLS handshake protocol Finished message.

    .. note::
      In TLS 1.3 the message will modify record layer to start *sending*
      records with encryption using the ``client_handshake_traffic_secret``
      keys.
      It will also modify the record layer to start expecting the records
      to be encrypted with ``server_application_traffic_secret`` keys.
    """

    def __init__(self, version=None, description=None):
        """
        Initialize object.

        .. note::
            The ``description`` parameter MUST be specified
            as a keyword argument, i.e. read the definition as
            ``(self, *, description=None)`` (see PEP 3102).
            Otherwise the behaviour of this node is not guaranteed if new
            arguments are added to it (as they will be added *before*
            the ``description`` argument).

        :param str description: name or comment attached to the node,
            it will be printed when :py:func:`str` or :py:func:`repr` is
            called on the node.
        """
        if version in ((0, 2), (2, 0)):
            super(ExpectFinished, self).__init__(ContentType.handshake,
                                                 SSL2HandshakeType.
                                                 server_finished)
        else:
            super(ExpectFinished, self).__init__(ContentType.handshake,
                                                 HandshakeType.finished)
        self.version = version
        self.description = description

    def process(self, state, msg):
        """
        :type state: ConnectionState
        :type msg: Message
        """
        assert msg.contentType == ContentType.handshake
        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == self.handshake_type
        if self.version is None:
            self.version = state.version

        if self.version in ((0, 2), (2, 0)):
            finished = ServerFinished()
        else:
            finished = Finished(self.version, state.prf_size)

        finished.parse(parser)

        if self.version in ((0, 2), (2, 0)):
            state.session_id = finished.verify_data
        elif self.version <= (3, 3):
            verify_expected = calc_key(state.version,
                                       state.key['master_secret'],
                                       state.cipher,
                                       b'client finished' if not state.client
                                       else b'server finished',
                                       state.handshake_hashes,
                                       output_length=12)

            assert finished.verify_data == verify_expected
        else:  # TLS 1.3
            finished_key = HKDF_expand_label(
                state.key['server handshake traffic secret'],
                b'finished',
                b'',
                state.prf_size,
                state.prf_name)
            transcript_hash = state.handshake_hashes.digest(state.prf_name)
            verify_expected = secureHMAC(finished_key,
                                         transcript_hash,
                                         state.prf_name)
            assert finished.verify_data == verify_expected

        state.handshake_messages.append(finished)
        state.key['server_verify_data'] = finished.verify_data
        state.handshake_hashes.update(msg.write())

        if self.version in ((0, 2), (2, 0)):
            state.msg_sock.handshake_finished = True

        if self.version > (3, 3):
            # in TLS 1.3 ChangeCipherSpec is a no-op, so we need to attach
            # the change for reading to some message that is always sent
            state.msg_sock.changeWriteState()

            # we now need to calculate application traffic keys to allow
            # correct interpretation of the alerts regarding Certificate,
            # CertificateVerify and Finished

            # derive the master secret
            secret = derive_secret(
                state.key['handshake secret'], b'derived', None,
                state.prf_name)
            secret = secureHMAC(
                secret, bytearray(state.prf_size), state.prf_name)
            state.key['master secret'] = secret

            # derive encryption keys
            c_traff_sec = derive_secret(
                secret, b'c ap traffic', state.handshake_hashes,
                state.prf_name)
            state.key['client application traffic secret'] = c_traff_sec
            s_traff_sec = derive_secret(
                secret, b's ap traffic', state.handshake_hashes,
                state.prf_name)
            state.key['server application traffic secret'] = s_traff_sec

            # derive TLS exporter key
            exp_ms = derive_secret(secret, b'exp master',
                                   state.handshake_hashes,
                                   state.prf_name)
            state.key['exporter master secret'] = exp_ms

            # set up the encryption keys for application data
            state.msg_sock.calcTLS1_3PendingState(
                state.cipher, c_traff_sec, s_traff_sec, None)
            state.msg_sock.changeReadState()

    def __repr__(self):
        """Return human readable representation of the object."""
        return self._repr(['description'])


class ExpectEncryptedExtensions(_ExpectExtensionsMessage):
    """Processing of the TLS handshake protocol Encrypted Extensions message"""

    def __init__(self, extensions=None):
        super(ExpectEncryptedExtensions, self).__init__(
            ContentType.handshake,
            HandshakeType.encrypted_extensions,
            extensions)

    def _compare_extensions_in_ee(self, srv_exts, cln_hello):
        """
        Verify that server provided extensions match exactly expected list.
        """
        # check if received extensions match the set extensions
        self._compare_extensions(srv_exts)
        if self.extensions is None and srv_exts.extensions:
            cln_exts = set(i.extType for i in cln_hello.extensions)
            got = set(i.extType for i in srv_exts.extensions)
            diff = got.difference(cln_exts)
            if not got.issubset(cln_exts):
                raise AssertionError("Server sent unexpected extension(s):"
                                     " {0}".format(
                                         ", ".join(ExtensionType.toStr(i)
                                                   for i in diff)))

    @staticmethod
    def _get_autohandler(ext_id):
        try:
            return _EE_EXT_HANDLER[ext_id]
        except KeyError:
            raise ValueError("No autohandler for "
                             "{0}"
                             .format(ExtensionType
                                     .toStr(ext_id)))

    def _process_extensions(self, state, srv_exts):
        """Check if extensions are correct."""
        # fix these constants, when the extensions are implemented
        ee_supported = [ExtensionType.server_name,
                        1,  # max_fragment_length - RFC 6066
                        ExtensionType.supported_groups,
                        14,  # use_srtp - RFC 5764
                        ExtensionType.heartbeat,  # RFC 6520
                        ExtensionType.alpn,
                        19,  # client_certificate_type
                             # draft-ietf-tls-tls13-28 / RFC 7250
                        20,  # server_certificate_type
                             # draft-ietf-tls-tls13-28 / RFC 7250
                        ExtensionType.record_size_limit,  # RFC 8449
                        ExtensionType.early_data]

        for ext in srv_exts.extensions:
            ext_id = ext.extType
            if ext_id not in ee_supported:
                raise AssertionError("Server sent unsupported "
                                     "extension of type {0}"
                                     .format(ExtensionType
                                             .toStr(ext_id)))
            handler = None
            if self.extensions:
                handler = self.extensions[ext_id]

            # use automatic handlers for some extensions
            if handler is None:
                handler = self._get_autohandler(ext_id)

            if callable(handler):
                handler(state, ext)
            elif isinstance(handler, TLSExtension):
                if not handler == ext:
                    raise AssertionError("Expected extension not "
                                         "matched for type {0}, "
                                         "received: {1}"
                                         .format(ExtensionType
                                                 .toStr(ext_id),
                                                 ext))
            else:
                raise ValueError("Bad extension handler for id {0}"
                                 .format(ExtensionType.toStr(ext_id)))

    def process(self, state, msg):
        assert msg.contentType == ContentType.handshake
        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == self.handshake_type

        srv_exts = EncryptedExtensions().parse(parser)

        # get client_hello message with CH extensions
        cln_hello = state.get_last_message_of_type(ClientHello)

        self._compare_extensions_in_ee(srv_exts, cln_hello)

        if srv_exts.extensions:
            self._process_extensions(state, srv_exts)

        state.handshake_messages.append(srv_exts)
        state.handshake_hashes.update(msg.write())


class ExpectNewSessionTicket(ExpectHandshake):
    """Processing TLS handshake protocol new session ticket message."""

    def __init__(self, version=None, description=None):
        """
        Initialise object.

        .. note::
            The ``description`` parameter MUST be specified
            as a keyword argument, i.e. read the definition as
            ``(self, *, description=None)`` (see PEP 3102).
            Otherwise the behaviour of this node is not guaranteed if new
            arguments are added to it (as they will be added *before*
            the ``description`` argument).

        :param tuple version: parse the message as in the specified TLS
            version, use negotiated version by default
        :param str description: name or comment attached to the node,
            it will be printed when :py:func:`str` or :py:func:`repr` is
            called on the node.
        """
        super(ExpectNewSessionTicket, self).__init__(
            ContentType.handshake,
            HandshakeType.new_session_ticket)
        self.description = description
        self.version = version

    def process(self, state, msg):
        """Parse, verify and process the message."""
        assert msg.contentType == ContentType.handshake
        msg_bytes = msg.write()
        parser = Parser(msg_bytes)
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.new_session_ticket
        if self.version is None:
            self.version = state.version

        if self.version < (3, 4):
            ticket = NewSessionTicket1_0().parse(parser)
        else:
            ticket = NewSessionTicket().parse(parser)
        ticket.time = time.time()

        state.session_tickets.append(ticket)

        if self.version < (3, 4):
            # in TLS 1.2 and earlier tickets are part of the Handshake, so
            # they need to be hashed
            state.handshake_messages.append(ticket)
            state.handshake_hashes.update(msg_bytes)

    def __repr__(self):
        """Return human readable representation of object."""
        return self._repr(['description'])


class ExpectHelloRequest(ExpectHandshake):
    """Processing of TLS handshake protocol hello request message."""

    def __init__(self, description=None):
        """
        Initialise object.

        .. note::
            The ``description`` parameter MUST be specified
            as a keyword argument, i.e. read the definition as
            ``(self, *, description=None)`` (see PEP 3102).
            Otherwise the behaviour of this node is not guaranteed if new
            arguments are added to it (as they will be added *before*
            the ``description`` argument).

        :param str description: name or comment attached to the node,
            it will be printed when :py:func:`str` or :py:func:`repr` is
            called on the node.
        """
        super(ExpectHelloRequest, self).__init__(
            ContentType.handshake,
            HandshakeType.hello_request)
        self.description = description

    def process(self, state, msg):
        """Parse, verify and process the message."""
        assert msg.contentType == ContentType.handshake
        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.hello_request

        # check if it is well-formed
        HelloRequest().parse(parser)

    def __repr__(self):
        """Return human readable representation of object."""
        return self._repr(['description'])


class ExpectAlert(Expect):
    """Processing TLS Alert message"""

    def __init__(self, level=None, description=None):
        super(ExpectAlert, self).__init__(ContentType.alert)
        self.level = level
        self.description = description

    def process(self, state, msg):
        assert msg.contentType == ContentType.alert
        parser = Parser(msg.write())

        alert = Alert()
        alert.parse(parser)

        problem_desc = ""
        if self.level is not None and alert.level != self.level:
            problem_desc += "Alert level {0} != {1}".format(alert.level,
                                                            self.level)
        if self.description is not None:
            # allow for multiple choice for description
            if not isinstance(self.description, Iterable):
                self.description = tuple([self.description])

            if alert.description not in self.description:
                if problem_desc:
                    problem_desc += ", "
                descriptions = ["\"{0}\"".format(AlertDescription.toStr(i))
                                for i in self.description]
                expected = ", ".join(
                    itertools.chain((i for i in descriptions[:-2]),
                                    [" or ".join(i for i in descriptions[-2:])]
                                   ))
                received = AlertDescription.toStr(alert.description)
                problem_desc += ("Expected alert description {0} does not "
                                 "match received \"{1}\""
                                 .format(expected, received))
        if problem_desc:
            raise AssertionError(problem_desc)

    def __repr__(self):
        """Return human readable representation of object."""
        return self._repr(["level", "description"])


class ExpectSSL2Alert(ExpectHandshake):
    """Processing of SSLv2 Handshake protocol alert messages"""

    def __init__(self, error=None):
        super(ExpectSSL2Alert, self).__init__(ContentType.handshake,
                                              SSL2HandshakeType.error)
        self.error = error

    def process(self, state, msg):
        """Analyse the error message"""
        assert msg.contentType == ContentType.handshake

        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == SSL2HandshakeType.error

        if self.error is not None:
            assert self.error == parser.get(2)


class ExpectApplicationData(Expect):
    """Processing Application Data message"""

    def __init__(self, data=None, size=None, output=None, description=None):
        super(ExpectApplicationData, self).\
                __init__(ContentType.application_data)
        self.data = data
        self.size = size
        self.output = output
        self.description = description

    def __str__(self):
        """Return human readable representation of the object."""
        return self._repr(['data', 'size', 'description'])

    def process(self, state, msg):
        assert msg.contentType == ContentType.application_data
        data = msg.write()

        if self.data:
            assert self.data == data
        if self.size and len(data) != self.size:
            raise AssertionError("ApplicationData of unexpected size: {0}, "
                                 "expected: {1}".format(len(data), self.size))
        if self.output:
            self.output.write("ExpectApplicationData received payload:\n")
            self.output.write(repr(data))
            self.output.write("ExpectApplicationData end of payload.\n")


class ExpectHeartbeat(ExpectMessage):
    """Processing of heartbeat messages."""

    def __init__(self, message_type=HeartbeatMessageType.heartbeat_response,
                 payload=None, padding_size=None):
        """
        Set up waiting for a heartbeat message.

        :type message_type: int
        :param message_type: Type of heartbeat messages to wait for, see
            `~tlslite.constants.HeartbeatMessageType` for defined types
        :type payload: bytes-like
        :param payload: literal value of padding to expect, if set to ``None``,
            any payload will be accepted
        :type padding_size: int
        :param padding_size: exact length of padding that will be expected,
            if set to ``None``, any padding length will be accepted
        """
        super(ExpectHeartbeat, self).\
            __init__(ContentType.heartbeat)
        self.message_type = message_type
        self.payload = payload
        self.padding_size = padding_size

    def process(self, state, msg):
        """Check if the ``msg`` meets the requirements for the message."""
        assert msg.contentType == ContentType.heartbeat

        parser = Parser(msg.write())
        heartbeat = Heartbeat().parse(parser)

        self._cmp_eq(self.message_type, heartbeat.message_type,
                     HeartbeatMessageType,
                     "Unexpected heartbeat message type. Expected: {0}, "
                     "received: {1}.")

        self._cmp_eq(self.payload, heartbeat.payload,
                     f_str="Unexpected payload in Heartbeat message "
                           "received. Expected: {0!r}, received: {1!r}")

        if self.padding_size is None:
            assert len(heartbeat.padding) >= 16
        else:
            if len(heartbeat.padding) != self.padding_size:
                raise AssertionError(
                        "Server sent unexpected size of padding "
                        "in heartbeat message. Expected: {0}, "
                        "received: {1}".format(self.padding_size,
                                               len(heartbeat.padding)))


class ExpectNoMessage(Expect):
    """
    Virtual message signifying timeout on message listen.

    :ivar timeout: how long to wait for message before giving up, in seconds,
        can be float
    :vartype timeout: int or float
    """

    def __init__(self, timeout=0.1):
        super(ExpectNoMessage, self).__init__(None)
        self.timeout = timeout

    def process(self, state, msg):
        """Do nothing."""
        pass


class ExpectClose(Expect):
    """Virtual message signifying closing of TCP connection"""

    def __init__(self):
        super(ExpectClose, self).__init__(None)

    def process(self, state, msg):
        """Close our side"""
        state.msg_sock.sock.close()


class ExpectCertificateStatus(ExpectHandshake):
    """Processing of CertificateStatus message from RFC 6066."""

    def __init__(self):
        super(ExpectCertificateStatus,
              self).__init__(ContentType.handshake,
                             HandshakeType.certificate_status)

    def process(self, state, msg):
        assert msg.contentType == ContentType.handshake

        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == HandshakeType.certificate_status

        cert_status = CertificateStatus().parse(parser)

        state.handshake_messages.append(cert_status)
        state.handshake_hashes.update(msg.write())


class ExpectKeyUpdate(ExpectHandshake):
    """Processing of post-handshake KeyUpdate message from RFC 8446"""

    def __init__(self, message_type=None):
        """
        Initialize object.

        :type message_type: int
        :param message_type: type of KeyUpdate msg, either
            update_not_requested or update_requested
        """
        super(ExpectKeyUpdate, self).__init__(
            ContentType.handshake,
            HandshakeType.key_update)
        self.message_type = message_type

    def process(self, state, msg):
        """
        Parse, verify and process the message.

        :type state: ConnectionState
        :type msg: Message
        """
        assert msg.contentType == self.content_type
        parser = Parser(msg.write())
        hs_type = parser.get(1)
        assert hs_type == self.handshake_type

        keyupdate = KeyUpdate().parse(parser)
        assert keyupdate.message_type == self.message_type

        _, sr_app_secret = state.msg_sock.\
            calcTLS1_3KeyUpdate_sender(
                state.cipher,
                state.key['client application traffic secret'],
                state.key['server application traffic secret'])
        state.key['server application traffic secret'] = sr_app_secret



# Author: Jan Koscielniak, (c) 2020
# Released under Gnu GPL v2.0, see LICENSE file for details

"""Extraction and analysis of timing information from a packet capture."""

from __future__ import print_function

import getopt
import sys
import csv
import time
import math
from os import remove
from os.path import join, splitext, getsize, exists
from collections import defaultdict
from socket import inet_aton, gethostbyname, gaierror, error
import multiprocessing as mp
from threading import Thread, Event
import hashlib
import tempfile
from random import choice
import ecdsa
import pandas as pd
import numpy as np

import dpkt

from tlsfuzzer.utils.log import Log
from tlsfuzzer.utils.statics import WARM_UP
from tlsfuzzer.utils.lists import natural_sort_keys
from tlsfuzzer.utils.ordered_dict import OrderedDict
from tlsfuzzer.utils.progress_report import progress_report
from tlsfuzzer.utils.compat import bit_count
from tlslite.utils.cryptomath import bytesToNumber, numberToByteArray
from tlslite.utils.python_key import Python_Key

try:
    from itertools import izip
except ImportError: # will be 3.x series
    izip = zip

if sys.version_info >= (3, 10):
    def bit_count(n):
        return n.bit_count()
else:
    def bit_count(n):
        return bin(n).count("1")

WAIT_FOR_FIRST_BARE_MAX_VALUE = 0
WAIT_FOR_NON_BARE_MAX_VALUE = 1
WAIT_FOR_SECOND_BARE_MAX_VALUE = 2


def help_msg():
    """Print help message."""
    print("Usage: extract [-l logfile] [-c capture] [[-o output] ...]")
    print(" -l logfile     Filename of the timing log (required)")
    print(" -c capture     Packet capture of the test run")
    print(" -o output      Directory where to place results (required)")
    print(" -h host        TLS server host or ip")
    print(" -p port        TLS server port")
    print(" -n name        column name to use from the raw-times file")
    print(" --raw-times FILE Read the timings from an external file, not")
    print("                the packet capture.")
    print(" --binary num   Expect the raw-times file to store binary numbers")
    print("                'num' bytes each. Note: using it will overwrite")
    print("                the csv counterpart to FILE (if FILE is 'data.bin'")
    print("                it will overwrite 'data.csv'")
    print(" --endian endian What endianness to use, 'little' or 'big', with")
    print("                little being the default")
    print(" --no-quickack  Don't assume QUICKACK to be in use (affects")
    print("                capture file parsing only)")
    print(" --status-delay num How often to print the status line.")
    print(" --status-newline Use newline instead of carriage return for")
    print("                printing status line. Useful when redirecting")
    print("                standard output to a file.")
    print(" --raw-data FILE Read the data used for processing from an")
    print("                external file. The file must be in binary format.")
    print("                Used for ECDSA and ECDH data processing.")
    print(" --data-size num The size of each data used for processing.")
    print(" --prehashed    Specifies that the data on the file are already")
    print("                hashed. Canceled by hash-func option.")
    print("                Used for ECDSA data processing.")
    print(" --raw-sigs FILE Read the signatures from an external file.")
    print("                The file must be in binary format.")
    print("                Used for ECDSA data processing.")
    print(" --sig-format [DER|RAW] Specifies the format of the signatures in")
    print("                the binary file, 'DER' or 'RAW'. Default DER.")
    print("                Used for ECDSA data processing.")
    print(" --raw-values FILE Read the values from an external file.")
    print("                The file must be in binary format.")
    print("                Used for ECDH timing data.")
    print(" --value-size num Override the size of each value. By default the")
    print("                script will try to calculate it.")
    print(" --value-endianness endian What endianness to use for the")
    print("                interpretation of the values, 'little' or 'big',")
    print("                with big being the default")
    print(" --priv-key-ecdsa FILE Read the ecdsa private key from PEM file.")
    print(" --clock-frequency freq Assume that the times in the file are not")
    print("                specified in seconds but rather in clock cycles of")
    print("                a clock running at frequency 'freq' specified in")
    print("                MHz. Use when the clock source are the raw reads")
    print("                from the Time Stamp Counter register or similar.")
    print(" --hash-func func Specifies the hash function to use for")
    print("                extracting the k value. The function should be")
    print("                available in hashlib module. The default function")
    print("                is sha256.")
    print(" --skip-invert  Skipping the creation of inverse of K value")
    print("                measurement files.")
    print("                ECDSA analysis only")
    print(" --rsa-keys FILE Analyse the times based on RSA private keys")
    print("                values. Creates separate measurements.csv files")
    print("                for the d, p, q, dP, dQ, and qInv values.")
    print("                Contents must be concatenated PKCS#8 PEM keys.")
    print(" --workers num  Number of worker processes to use for")
    print("                parallelizable computation. More workers")
    print("                will finish analysis faster, but will require")
    print("                more memory to do so. By default: number of")
    print("                threads available on the system (`os.cpu_count()`)")
    print(" --max-bit-size num Override the max bit size used in the creation")
    print("                of the tuples. By default the script will try to")
    print("                calculate it. Used only in the bit size extraction")
    print(" --verbose      Print's a more verbose output.")
    print(" --help         Display this message")
    print("")
    print("'logfile' is the file that specifies the order in which probes")
    print("were executed (usually named 'log.csv').")
    print("")
    print("When extracting data from a packet capture file, specifying the")
    print("capture file, host and port is necessary.")
    print("")
    print("When using the external timing source, only it, and the always")
    print("required options: logfile and output dir are necessary.")
    print("")
    print("When doing signature extraction, data file, data size, signatures")
    print("file, and one private key are necessary.")
    print("")
    print("For ECDSA signatures, the hash function used for the extraction of")
    print("K depends on the private key's curve size.")
    print("")
    print("When doing ECDH secret extraction, data file, data size, values")
    print("file, and one private key are necessary.")
    print("")
    print("When doing RSA key extraction, RSA private keys, raw times,")
    print("and output dir are necessary.")


def main():
    """Process arguments and start extraction."""
    logfile = None
    capture = None
    output = None
    ip_address = None
    port = None
    raw_times = None
    col_name = None
    binary = None
    endian = 'little'
    no_quickack = False
    delay = None
    carriage_return = None
    data = None
    data_size = None
    prehashed = False
    sigs = None
    sig_format = "DER"
    values = None
    value_size = None
    value_endianness = 'big'
    priv_key = None
    key_type = None
    freq = None
    hash_func_name = None
    invert = True
    rsa_keys = None
    workers = None
    max_bit_size = None
    verbose = False

    argv = sys.argv[1:]

    if not argv:
        help_msg()
        sys.exit(1)

    opts, args = getopt.getopt(argv, "l:c:h:p:o:t:n:",
                               ["help", "raw-times=", "binary=", "endian=",
                                "no-quickack", "status-delay=",
                                "status-newline", "raw-data=", "data-size=",
                                "prehashed", "raw-sigs=", "sig-format=",
                                "raw-values=", "value-size=",
                                "value-endianness=", "priv-key-ecdsa=",
                                "clock-frequency=", "hash-func=",
                                "skip-invert", "workers=", "rsa-keys=",
                                "max-bit-size=", "verbose"])
    for opt, arg in opts:
        if opt == '-l':
            logfile = arg
        elif opt == '-c':
            capture = arg
        elif opt == '-o':
            output = arg
        elif opt == '-h':
            ip_address = arg
        elif opt == "-n":
            col_name = arg
        elif opt == '-p':
            port = int(arg)
        elif opt == "--raw-times":
            raw_times = arg
        elif opt == "--binary":
            binary = int(arg)
        elif opt == "--endian":
            endian = arg
        elif opt == "--no-quickack":
            no_quickack = True
        elif opt == "--status-delay":
            delay = float(arg)
        elif opt == "--status-newline":
            carriage_return = '\n'
        elif opt == "--raw-data":
            data = arg
        elif opt == "--data-size":
            data_size = int(arg)
        elif opt == "--prehashed":
            prehashed = True
        elif opt == "--raw-sigs":
            sigs = arg
        elif opt == "--sig-format":
            sig_format = arg
        elif opt == "--raw-values":
            values = arg
        elif opt == "--value-size":
            value_size = int(arg)
        elif opt == "--value-endianness":
            value_endianness = arg
        elif opt == "--rsa-keys":
            rsa_keys = arg
        elif opt == "--priv-key-ecdsa":
            priv_key = arg
            if not key_type:
                key_type = "ec"
            else:
                raise ValueError(
                    "Can't specify more than one private key.")
        elif opt == "--verbose":
            verbose = True
        elif opt == "--clock-frequency":
            freq = float(arg) * 1e6
        elif opt == "--hash-func":
            hash_func_name = arg
        elif opt == "--skip-invert":
            invert = False
        elif opt == "--workers":
            workers = int(arg)
        elif opt == "--max-bit-size":
            max_bit_size = int(arg)
        elif opt == "--help":
            help_msg()
            sys.exit(0)

    if args:
        raise ValueError(
            "Unexpected arguments: {0}".format(args))

    if raw_times and capture:
        raise ValueError(
            "Can't specify both a capture file and external timing log")

    if binary and col_name:
        raise ValueError(
            "Binary format doesn't support column names")

    if binary and not raw_times:
        raise ValueError(
            "Can't specify binary number size without raw-times file")

    if endian not in ('little', 'big'):
        raise ValueError(
            "Only 'little' and 'big' endianess supported")

    if not all([any([logfile, sigs, rsa_keys, values]), output]):
        raise ValueError(
            "Specifying either logfile, rsa keys, raw sigs or raw values "
            "and output is mandatory")

    if capture and not all([logfile, output, ip_address, port]):
        raise ValueError("Some arguments are missing!")

    if sigs and not all([raw_times, data, data_size, sigs, priv_key]):
        raise ValueError(
            "When doing signature extraction, times file, data file, "
            "data size, signatures file and one private key are necessary.")

    if values and not all([values, priv_key, raw_times, data]):
        raise ValueError(
            "When doing ECDH secret extraction, times file, data file, "
            "secrets file, and one private key are necessary.")

    if hash_func_name == None:
        if prehashed:
            hash_func = None
        else:
            hash_func = hashlib.sha256
    else:
        try:
            hash_func = getattr(hashlib, hash_func_name)
        except AttributeError:
            raise ValueError(
                "Hash function {0} is not supported.".format(hash_func_name))

    log = None
    if logfile:
        log = Log(logfile)
        log.read_log()

    extract = Extract(
        log, capture, output, ip_address, port, raw_times, col_name,
        binary=binary, endian=endian, no_quickack=no_quickack,
        delay=delay, carriage_return=carriage_return,
        data=data, data_size=data_size, sigs=sigs, priv_key=priv_key,
        key_type=key_type, frequency=freq, hash_func=hash_func,
        workers=workers, verbose=verbose, rsa_keys=rsa_keys,
        sig_format=sig_format, values=values, value_size=value_size,
        value_endianness=value_endianness, max_bit_size=max_bit_size
    )
    extract.parse()

    if any([sigs, values]) and all([raw_times, data, priv_key]):
        files = {
            "measurements.csv": "k-size" if not values else "size",
            "measurements-hamming-weight.csv": "hamming-weight"
        }

        if invert and not values:
            file_list = list(files.keys())

            for file in file_list:
                invert_file_name = file.split(".")[0] + '-invert.csv'
                files[invert_file_name] = "invert-" + files[file]

        extract.process_and_create_multiple_csv_files(
            files, ecdh=(values is not None))

    if rsa_keys:
        extract.process_rsa_keys()


class Extract:
    """Extract timing information from packet capture."""

    def __init__(self, log=None, capture=None, output=None, ip_address=None,
                 port=None, raw_times=None, col_name=None,
                 write_csv='timing.csv', write_pkt_csv='raw_times_detail.csv',
                 measurements_csv="measurements.csv",
                 binary=None, endian='little', no_quickack=False, delay=None,
                 carriage_return=None, data=None, data_size=None, sigs=None,
                 priv_key=None, key_type=None, frequency=None,
                 hash_func=hashlib.sha256, workers=None, verbose=False,
                 fin_as_resp=False, rsa_keys=None, sig_format="DER",
                 values=None, value_size=None, value_endianness="little",
                 max_bit_size=None):
        """
        Initialises instance and sets up class name generator from log.

        :param Log log: Log class instance
        :param str capture: Packet capture filename
        :param str output: Directory where to output results
        :param str ip_address: TLS server ip address
        :param int port: TLS server port
        :param str data: Name of file with data used for processing
        :param int data_size: Size of each data used for processing
        :param str sigs: Name of file with the signatures
        :param str sig_format: The format the signatures have, DER and RAW
            formats are supported
        :param str values: Name of file with the values
        :param int values_size: Size of each value
        :param int values_endianness: endianess for the interpretation of the
            values
        :param str priv_key: Private key filename
        :param str key_type: The type of the private key, for now only "ec" is
            supported
        :param int binary: Number of bytes per timing from raw times file
        :param str endian: Endianess of the read numbers
        :param bool no_quickack: If True, don't expect QUICKACK to be in use
        :param float delay: How often to print the status line.
        :param str carriage_return: What chacarter to use as status line end.
        :param func hash_func: The hash function that will be used for hashing
            the message in bit size analysis. None for prehashed data.
        :param int workers: The amount of parallel workers to be used.
        :param bool verbose: Prints a more verbose output
        :param bool fin_as_resp: Consider the server FIN packet to be the
            response to previous client query
        """
        self.capture = capture
        self.output = output
        self.ip_address = ip_address and self.hostname_to_ip(ip_address)
        self.port = port
        self.timings = defaultdict(list)
        self.client_message = None
        self.server_message = None
        self.client_msgs = []
        self.server_msgs = []
        self.initial_syn = None
        self.initial_syn_ack = None
        self.initial_ack = None
        self.warm_up_messages_left = WARM_UP
        self.last_warmup_fin = None
        self.raw_times = raw_times
        self.binary = binary
        self.endian = endian
        self.pckt_times = []
        self.col_name = col_name
        self.write_csv = write_csv
        self.write_pkt_csv = write_pkt_csv
        self._exp_clnt = None
        self._exp_srv = None
        self._previous_lst_msg = None
        self._write_class_names = None
        self.no_quickack = no_quickack
        self.delay = delay
        self.carriage_return = carriage_return
        self.data = data
        self.data_size = data_size
        self.sigs = sigs
        self.r_or_s_size = None
        self.frequency = frequency
        self.measurements_csv = measurements_csv
        self.hash_func = hash_func  # None if data are already hashed
        self.workers = workers
        self.verbose = verbose
        self._total_measurements = None
        self._measurements_fp = None
        self._intermedian_fp = None
        self._current_line = None
        self._next_line = None
        self._line_to_write = None
        self._max_tuple_size = 0
        self._measurements_dropped = 0
        self._selections = None
        self._row = 0
        self._max_value = None
        self._fin_as_resp = fin_as_resp
        self.rsa_keys = rsa_keys
        self._temp_HWI_name = None
        self.values = values
        self.value_size = value_size
        self.value_endianness = value_endianness
        self.max_bit_size = max_bit_size

        if sig_format not in ["DER", "RAW"]:
            raise ValueError(
                "Unknown signature format {0}. ".format(sigs_format) +
                "Please use 'DER' or 'RAW'."
            )

        if data and data_size:
            try:
                self._total_measurements = int(getsize(data) / data_size)
            except OSError:
                self._total_measurements = None

        self.priv_key = None
        if key_type == "ec":
            with open(priv_key, 'r') as f:
                self.priv_key = ecdsa.SigningKey.from_pem(f.read())
            if sig_format == 'RAW':
                self.r_or_s_size = self.priv_key.baselen

        # set up class names generator
        self.log = log
        if log:
            self.class_generator = log.iterate_log()
            self.class_names = log.get_classes()

    def parse(self):
        """
        Extract timing information from capture file
        and associate it with class from log file.
        """
        if self.capture:
            return self._parse_pcap()
        return self._parse_raw_times()

    def _convert_binary_file(self, raw_times_name):
        """Convert the binary file format to csv before further processing."""
        times_iter = self._get_time_from_file()

        with open(raw_times_name, 'w') as raw_times:
            raw_times.write("raw times\n")
            for val in times_iter:
                raw_times.write(str(val) + '\n')

    def _parse_raw_times(self):
        """Classify already extracted times."""
        # as unlike with capture file, we don't know how many sanity tests,
        # manual checks, etc. were performed to the server before the
        # timing tests were started, we don't know how many measurements to
        # skip. Count the probes, the times, and then use the last len(probes)
        # of times for classification

        raw_times_name = self.raw_times

        # if we got a binary file on input, first transform it into a csv file
        if self.binary:
            raw_times_name = splitext(self.raw_times)[0] + ".csv"
            self._convert_binary_file(raw_times_name)

        if not self.log:
            return

        # do counting in memory efficient way
        probe_count = sum(1 for _ in self.class_generator)
        self.log.read_log()
        self.class_generator = self.log.iterate_log()
        with open(raw_times_name, 'r') as raw_times:
            # skip the header line
            raw_times.readline()
            times_count = 0
            for times_count, _ in enumerate(raw_times, 1):
                pass
        if probe_count > times_count:
            raise ValueError(
                "Insufficient number of times for provided log file "
                "(expected: {0}, found: {1})".format(probe_count, times_count))

        self.warm_up_messages_left = times_count - probe_count

        data_iter = self._get_time_from_file()
        for _ in range(self.warm_up_messages_left):
            next(data_iter)

        for line in data_iter:
            class_index = next(self.class_generator)
            class_name = self.class_names[class_index]
            self.timings[class_name].append(line)
            self._flush_to_files()

        self._write_csv_header()
        self._write_csv()

    def _parse_pcap(self):
        """Process capture file."""
        with open(self.capture, 'rb') as pcap:
            progress = None
            try:
                pcap.seek(0, 2)
                exp_len = pcap.tell()
                pcap.seek(0, 0)
                status = [0, exp_len, Event()]
                kwargs = {}
                kwargs['unit'] = 'B'
                kwargs['prefix'] = 'binary'
                kwargs['delay'] = self.delay
                kwargs['end'] = self.carriage_return
                progress = Thread(target=progress_report, args=(status,),
                                  kwargs=kwargs)
                progress.start()

                capture = dpkt.pcap.Reader(pcap)

                exp_srv_ack = 0
                exp_clnt_ack = 0

                pkt_count = 0
                # since timestamp is Decimal() we don't have to worry about
                # float() precision
                for timestamp, pkt in capture:
                    status[0] = pcap.tell()
                    pkt_count += 1
                    link_packet = dpkt.ethernet.Ethernet(pkt)
                    ip_pkt = link_packet.data
                    tcp_pkt = ip_pkt.data

                    if (tcp_pkt.flags & dpkt.tcp.TH_SYN and
                            tcp_pkt.dport == self.port and
                            ip_pkt.dst == self.ip_address):
                        # a SYN packet was found - new connection
                        # (if a retransmission it won't be counted as at least
                        # one client and one server message has to be
                        # exchanged)
                        self.add_timing()

                        # reset timestamps
                        self.server_message = None
                        self.client_message = None
                        self.initial_syn = timestamp
                        self.initial_syn_ack = None
                        self.initial_ack = None
                        self.client_msgs = []
                        self.client_msgs_acks = OrderedDict()
                        self.server_msgs = []
                        self.server_msgs_acks = OrderedDict()
                        self.clnt_fin = None
                        self.srv_fin = None
                        self.ack_for_fin = None
                        self.in_srv_shutdown = False
                        self.in_clnt_shutdown = False
                        self.initial_ack_seq_no = None
                        exp_srv_ack = tcp_pkt.seq + 1 & 0xffffffff
                        exp_clnt_ack = 0
                    elif (tcp_pkt.flags & dpkt.tcp.TH_SYN and
                            tcp_pkt.flags & dpkt.tcp.TH_ACK and
                            tcp_pkt.sport == self.port and
                            ip_pkt.src == self.ip_address):
                        self.initial_syn_ack = timestamp
                        exp_clnt_ack = tcp_pkt.seq + 1 & 0xffffffff
                        if tcp_pkt.ack != exp_srv_ack:
                            print("Mismatched syn/ack seq at {0}\n"
                                  .format(pkt_count))
                            raise ValueError("Packet drops in capture!")
                    elif (tcp_pkt.flags & dpkt.tcp.TH_ACK and
                            tcp_pkt.dport == self.port and
                            ip_pkt.dst == self.ip_address and
                            tcp_pkt.ack == exp_clnt_ack and
                            not self.initial_ack):
                        # the initial ACK is the first ACK that acknowledges
                        # the SYN+ACK
                        self.initial_ack = timestamp
                        self.initial_ack_seq_no = tcp_pkt.ack
                    elif (tcp_pkt.flags & dpkt.tcp.TH_ACK and
                            not tcp_pkt.flags & dpkt.tcp.TH_FIN and
                            tcp_pkt.sport == self.port and
                            tcp_pkt.ack not in self.client_msgs_acks and
                            not self.in_srv_shutdown):
                        # check if it's the first ACK to a client sent message
                        if len(self.client_msgs) > len(self.client_msgs_acks):
                            self.client_msgs_acks[tcp_pkt.ack] = timestamp
                    elif (tcp_pkt.flags & dpkt.tcp.TH_ACK and
                            not tcp_pkt.flags & dpkt.tcp.TH_FIN and
                            tcp_pkt.dport == self.port and
                            tcp_pkt.ack != self.initial_ack_seq_no and
                            tcp_pkt.ack not in self.server_msgs_acks and
                            not self.in_clnt_shutdown):
                        # check if it's the first ACK to a server sent message
                        if len(self.server_msgs) > len(self.server_msgs_acks):
                            self.server_msgs_acks[tcp_pkt.ack] = timestamp
                    elif tcp_pkt.flags & dpkt.tcp.TH_FIN:
                        if tcp_pkt.sport == self.port:
                            self.in_srv_shutdown = True
                            self.srv_fin = timestamp
                            if len(self.client_msgs) > \
                                    len(self.client_msgs_acks):
                                self.client_msgs_acks[tcp_pkt.ack] = timestamp
                        else:
                            self.in_clnt_shutdown = True
                            self.clnt_fin = timestamp
                            if len(self.server_msgs) > \
                                    len(self.server_msgs_acks):
                                self.server_msgs_acks[tcp_pkt.ack] = timestamp
                    elif (tcp_pkt.flags & dpkt.tcp.TH_ACK and
                            not tcp_pkt.flags & dpkt.tcp.TH_FIN and
                            self.in_clnt_shutdown and self.in_srv_shutdown):
                        self.ack_for_fin = timestamp

                    # initial ACK can be combined with the first data packet
                    if tcp_pkt.data:
                        if (tcp_pkt.sport == self.port and
                                ip_pkt.src == self.ip_address):
                            if tcp_pkt.ack != exp_srv_ack:
                                print("Mismatched syn/ack seq at {0}\n"
                                      .format(pkt_count))
                                raise ValueError("Packet drops in capture!")
                            exp_clnt_ack = exp_clnt_ack + len(tcp_pkt.data) \
                                & 0xffffffff
                            # message from the server
                            self.server_message = timestamp
                            self.server_msgs.append(timestamp)
                        else:
                            if tcp_pkt.ack != exp_clnt_ack:
                                print("Mismatched syn/ack seq at {0}\n"
                                      .format(pkt_count))
                                raise ValueError("Packet drops in capture!")
                            exp_srv_ack = exp_srv_ack + len(tcp_pkt.data) \
                                & 0xffffffff
                            # message from the client
                            self.client_message = timestamp
                            self.client_msgs.append(timestamp)

                # deal with the last connection
                self.add_timing()
            finally:
                status[2].set()
                progress.join()
                print()

    def add_timing(self):
        """Associate the timing information with its class"""
        if self.client_message and self.server_message:
            if self.warm_up_messages_left == 0:
                class_index = next(self.class_generator)
                class_name = self.class_names[class_index]
                lst_clnt_ack = 0
                for lst_clnt_ack in self.client_msgs_acks.values():
                    pass
                if self._fin_as_resp:
                    srv_time = self.srv_fin
                else:
                    srv_time = self.server_msgs[-1]
                if self.no_quickack:
                    time_diff = srv_time - self.client_msgs[-1]
                else:
                    time_diff = srv_time - lst_clnt_ack
                self.timings[class_name].append(time_diff)
                self.pckt_times.append((
                    self.initial_syn,
                    self.initial_syn_ack,
                    self.initial_ack,
                    self.client_msgs,
                    self.client_msgs_acks,
                    self.server_msgs,
                    self.server_msgs_acks,
                    self.srv_fin,
                    self.clnt_fin,
                    self.ack_for_fin,
                ))
                self._flush_to_files()
            else:
                self.warm_up_messages_left -= 1
                if self.warm_up_messages_left == 0:
                    if self.srv_fin is None and self.clnt_fin is None:
                        self.last_warmup_fin = 0
                    else:
                        if self.srv_fin > self.clnt_fin:
                            self.last_warmup_fin = self.srv_fin
                        else:
                            self.last_warmup_fin = self.clnt_fin

    def _flush_to_files(self):
        # we can write only complete lines
        if len(self.timings) != len(self.class_names) or \
                not all(self.timings.values()):
            return

        if not self.raw_times:
            # make sure the csv has a header
            self._write_pkt_header()

            # then write queued up individual packet times
            self._write_pkts()

        # write the header of the already sorted results
        self._write_csv_header()

        # finally write the times of already sorted classes
        self._write_csv()

    def _write_csv_header(self):
        if self._write_class_names is not None:
            return

        filename = join(self.output, self.write_csv)
        with open(filename, 'w') as csvfile:
            print("Writing to {0}\n".format(filename))
            writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
            class_names = sorted(self.timings, key=natural_sort_keys)
            writer.writerow(class_names)
            self._write_class_names = class_names

    def _write_csv(self):
        filename = join(self.output, self.write_csv)
        with open(filename, 'a') as csvfile:
            writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
            for values in zip(*[self.timings[i] for i in
                    self._write_class_names]):
                writer.writerow("{0:.9e}".format(float(i)) for i in values)

            for i in self.timings.values():
                i.clear()

    def _write_pkts(self):
        for _, _, _, clnt_msgs, clnt_msgs_acks, srv_msgs, srv_msgs_acks, _, _, _ in self.pckt_times:
            if len(clnt_msgs) != len(clnt_msgs_acks): # pragma: no cover
                # no coverage; assert
                print(clnt_msgs)
                print()
                print(clnt_msgs_acks)
                raise ValueError("client message ACKs mismatch: {0} vs {1}"
                    .format(len(clnt_msgs), len(clnt_msgs_acks)))
            if len(srv_msgs) != len(srv_msgs_acks):  # pragma: no cover
                # no coverage; assert
                print(srv_msgs)
                print()
                print(srv_msgs_acks)
                raise ValueError("server message ACKs mismatch")

            if len(clnt_msgs) != self._exp_clnt:  # pragma: no cover
                # no coverage; assert
                raise ValueError("inconsistent count of client messages")

            if len(srv_msgs) + int(self._fin_as_resp) != self._exp_srv:  # pragma: no cover
                # no coverage: assert
                raise ValueError("inconsistent count of server messages")

        if self._previous_lst_msg is None:
            self._previous_lst_msg = self.last_warmup_fin

        filename = join(self.output, self.write_pkt_csv)
        with open(filename, "a") as csvfile:
            writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
            while self.pckt_times:
                (syn, syn_ack, ack, c_msgs, c_msgs_acks, s_msgs, s_msgs_acks,
                    srv_fin, clnt_fin, ack_for_fin) = self.pckt_times.pop(0)

                if self._previous_lst_msg is None:
                    row = [
                        0,
                        syn_ack - syn,
                        ack - syn_ack,
                    ]
                else:
                    row = [
                        syn - self._previous_lst_msg,
                        syn_ack - syn,
                        ack - syn_ack,
                    ]

                prv_ack = ack
                for c_msg, c_msg_ack, s_msg, s_msg_ack in zip(
                        c_msgs, c_msgs_acks.values(),
                        s_msgs, s_msgs_acks.values()
                ):
                    # prv_ack_to_clnt_X
                    row.append(c_msg - prv_ack)
                    prv_ack = s_msg_ack
                    # clnt_X_ack
                    row.append(c_msg_ack - c_msg)
                    # clnt_X_rtt
                    row.append(s_msg - c_msg)

                    # prv_ack_to_srv_X
                    row.append(s_msg - c_msg_ack)
                    # srv_X_ack
                    row.append(s_msg_ack - s_msg)

                # lst_srv_to_srv_fin
                if srv_fin is None:
                    row.append(0)
                else:
                    row.append(srv_fin - s_msgs[-1])

                # lst_srv_to_clnt_fin
                if clnt_fin is None:
                    row.append(0)
                else:
                    row.append(clnt_fin - s_msgs[-1])

                if srv_fin is None or clnt_fin is None:
                    row.append(0)
                else:
                    if srv_fin > clnt_fin:
                        last_fin = srv_fin
                    else:
                        last_fin = clnt_fin
                    # second_fin_to_ack
                    if ack_for_fin:
                        row.append(ack_for_fin - last_fin)
                        self._previous_lst_msg = ack_for_fin
                    else:
                        row.append(0.0)
                        self._previous_lst_msg = last_fin

                writer.writerow(row)

    def _write_pkt_header(self):
        if self._exp_clnt is not None:
            return

        for _, _, _, clnt_msgs, clnt_msgs_acks, srv_msgs, srv_msgs_acks, \
                srv_fin, _, _ in self.pckt_times:
            if len(clnt_msgs) != len(clnt_msgs_acks):  # pragma: no cover
                # no overage; assert
                print(clnt_msgs)
                print()
                print(clnt_msgs_acks)
                raise ValueError("client message ACKs mismatch: {0} vs {1}"
                    .format(len(clnt_msgs), len(clnt_msgs_acks)))
            if len(srv_msgs) != len(srv_msgs_acks):  # pragma: no cover
                # no coverage; assert
                print(srv_msgs)
                print()
                print(srv_msgs_acks)
                raise ValueError("server message ACKs mismatch")

            self._exp_clnt = len(clnt_msgs)
            self._exp_srv = len(srv_msgs) + int(self._fin_as_resp)

        if self._exp_srv != self._exp_clnt:  # pragma: no cover
            # no coverage; assert
            print(clnt_msgs)
            print(srv_msgs)
            raise ValueError("For every client query we need a response "
                             "(try FIN as server response)")

        filename = join(self.output, self.write_pkt_csv)
        with open(filename, 'w') as csvfile:
            print("Writing to {0}\n".format(filename))
            writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
            columns = [
                "lst_msg_to_syn",
                "syn_to_syn_ack",
                "syn_ack_to_ack",
            ]
            for i in range(self._exp_clnt):
                columns.append("prv_ack_to_clnt_{0}".format(i))
                columns.append("clnt_{0}_ack".format(i))
                columns.append("clnt_{0}_rtt".format(i))
                columns.append("prv_ack_to_srv_{0}".format(i))
                columns.append("srv_{0}_ack".format(i))
            columns.extend([
                "lst_srv_to_srv_fin",
                "lst_srv_to_clnt_fin",
                "second_fin_to_ack"
            ])
            writer.writerow(columns)

    def _get_data_from_binary_file(
            self, filename, data_size, convert_to_int=False, endian=None
        ):
        """
        Iterator. Reading raw bytes of data_size from a binary file. Can also
        convert the data to int.
        """
        if not endian:
            endian = self.endian

        with open(filename, "rb") as data_fp:
            data = data_fp.read(data_size)
            while data:
                if convert_to_int:
                    data = bytesToNumber(data, endian=endian)
                yield data
                data = data_fp.read(data_size)

    def _get_data_from_csv_file(self, filename, col_name=None,
                                convert_to_float=False, convert_to_int=False):
        """
        Iterator. Reading data from a csv file. Can also convert the data to
        float or to integer.
        """
        if not col_name:
            col_name = self.col_name

        with open(filename, "r") as data_fp:
            reader = csv.reader(data_fp)
            columns = next(reader)
            column = 0

            if len(columns) > 1 and col_name is None:
                raise ValueError(
                    "Multiple columns in {0} and ".format(filename) +
                    "no column name specified!"
                )

            if col_name:
                column = columns.index(col_name)

            for row in reader:
                data = row[column]
                if convert_to_float:
                    data = float(data)
                if convert_to_int:
                    data = int(data)
                yield data

    def _divide_by_frequency(self, value_iter):
        """Iterator. Devides value for given iter by frequency."""
        for value in value_iter:
            yield value / self.frequency

    def _get_time_from_file(self, filename=None):
        """Iterator. Read the times from file provided"""
        if self.binary:
            times_iter = self._get_data_from_binary_file(
                self.raw_times, filename if filename else self.binary,
                convert_to_int=True
            )
        else:
            times_iter = self._get_data_from_csv_file(
                filename if filename else self.raw_times, convert_to_float=True
            )

        if self.frequency:
            times_iter = self._divide_by_frequency(times_iter)

        return times_iter

    def _ecdsa_get_raw_signature_from_file_pointer(self, filename):
        """Iterator. Read the raw signatures from file provided"""
        with open(filename, "rb") as sigs_fp:
            r_and_s = sigs_fp.read(self.r_or_s_size * 2)
            while r_and_s:
                if len(r_and_s) != 2 * self.r_or_s_size:
                    raise ValueError(
                        "There was an error in parsing signatures. " +
                        "Incomplete r or s values in binary file.")
                yield r_and_s
                r_and_s = sigs_fp.read(self.r_or_s_size * 2)

    def _ecdsa_get_der_signature_from_file_pointer(self, filename):
        """Iterator. Read the DER signatures from file provided"""
        with open(filename, "rb") as sigs_fp:
            sig = sigs_fp.read(1)
            while sig:
                if not ecdsa.der.is_sequence(sig):
                    raise ValueError(
                        "There was an error in parsing signatures.")
                length_bytes = sigs_fp.read(1)
                sig_length = 0
                try:
                    sig_length = ecdsa.der.read_length(length_bytes)[0]
                except ecdsa.UnexpectedDER:
                    length_bytes += sigs_fp.read(1)
                    try:
                        sig_length = ecdsa.der.read_length(length_bytes)[0]
                    except ecdsa.UnexpectedDER:
                        raise ValueError(
                            "Couldn't read size of a signature.")
                sig_data = sigs_fp.read(sig_length)
                if sig_length != len(sig_data):
                    raise \
                        ValueError("Signature file ended unexpectedly.")
                sig += length_bytes + sig_data
                yield sig
                sig = sigs_fp.read(1)

    def _ecdsa_get_signature_from_file(self, filename=None):
        """Iterator. Read the signatures from file provided"""
        filename = filename if filename else self.sigs
        if self.r_or_s_size:
            return self._ecdsa_get_raw_signature_from_file_pointer(filename)
        else:
            return self._ecdsa_get_der_signature_from_file_pointer(filename)

    def _ecdsa_message_to_int(self, filename=None):
        """Iterator. Hashes the message used and converts it to int."""
        data_iter = self._get_data_from_binary_file(
            filename if filename else self.data, self.data_size
        )

        for msg in data_iter:
            if self.hash_func:
                hashed = self.hash_func(msg).digest()
                hashed = hashed[: self.priv_key.curve.baselen]
            else:
                hashed = msg
            number = int.from_bytes(hashed, 'big')
            max_length = ecdsa.util.bit_length(self.priv_key.curve.order)
            length = len(hashed) * 8
            number >>= max(0, length - max_length)
            yield number

    def _ecdsa_calculate_k(self, sig_and_hashed):
        """Iterator. Calculated the K value from a singature."""
        try:
            sig, hashed = sig_and_hashed
        except ValueError:
            raise ValueError(
                "Signature or hash not provided."
            )

        n_value = self.priv_key.curve.order
        g_value = self.priv_key.curve.generator

        if self.r_or_s_size:
            r_value, s_value = ecdsa.util.sigdecode_string(sig, n_value)
        else:
            r_value, s_value = ecdsa.util.sigdecode_der(sig, n_value)

        k_value = (
            (hashed + (
                r_value * self.priv_key.privkey.secret_multiplier
            ))
            * ecdsa.ecdsa.numbertheory.inverse_mod(s_value, n_value)
            ) % n_value
        kxg = (k_value * g_value).to_affine().x()

        if kxg == r_value:
            return k_value
        else:
            raise ValueError(
                "Failed to calculate k from given signatures.")

    def _convert_to_bit_size(self, value_iter):
        """Iterator. Convert a value to the bit length of it."""
        for value in value_iter:
            yield ecdsa.util.bit_length(value)

    def _convert_to_hamming_weight(self, value_iter):
        """Iterator. Converts a value to the Hamming weight of it."""
        for value in value_iter:
            yield bit_count(value)

    def _calculate_invert_k(self, value_iter):
        """Iterator. It will calculate the invert K."""
        n_value = self.priv_key.curve.order

        if self._temp_HWI_name and not exists(self._temp_HWI_name):
            with open(self._temp_HWI_name, "w") as fp:
                fp.write("{0}\n".format("invert_K_bit_count"))
                for value in value_iter:
                    invert = ecdsa.ecdsa.numbertheory.inverse_mod(
                        value, n_value)
                    invert_bit_count = bit_count(invert)
                    fp.write("{0}\n".format(invert_bit_count))
                    yield invert
        else:
            for value in value_iter:
                invert = ecdsa.ecdsa.numbertheory.inverse_mod(value, n_value)
                yield invert

    def ecdsa_iter(self, return_type="k-size"):
        """
        Iterator. Iterator to use for signatures signed by ECDSA private key.
        """
        k_map_filename = join(self.output, "ecdsa-k-time-map.csv")
        sigs_iter = self._ecdsa_get_signature_from_file()
        hashed_iter = self._ecdsa_message_to_int()
        times_iter = self._get_time_from_file()

        if not exists(k_map_filename):
            if self.verbose:
                print("[i] Creating ecdsa-k-time-map.csv file...")

            progress = None
            status = [0]
            if self.verbose and self._total_measurements:
                status = [0, self._total_measurements, Event()]
                kwargs = {}
                kwargs['unit'] = ' signatures'
                kwargs['prefix'] = 'decimal'
                kwargs['delay'] = self.delay
                kwargs['end'] = self.carriage_return
                progress = Thread(target=progress_report, args=(status,),
                                kwargs=kwargs)
                progress.start()

            try:
                with open(k_map_filename, "w") as fp:
                    with mp.Pool(self.workers) as pool:
                        k_iter = pool.imap(
                            self._ecdsa_calculate_k,
                            izip(sigs_iter, hashed_iter), 10000
                        )

                        fp.write("k_value,time\n")

                        for k_value, time_value in izip(k_iter, times_iter):
                            fp.write("{0},{1}\n".format(k_value, time_value))
                            status[0] += 1
            finally:
                if progress:
                    status[2].set()
                    progress.join()
                    print()

        k_iter = self._get_data_from_csv_file(
            k_map_filename, col_name="k_value", convert_to_int=True
        )

        if "invert" in return_type:
            if ("hamming-weight" in return_type and self._temp_HWI_name
                    and exists(self._temp_HWI_name)):
                return self._get_data_from_csv_file(
                    filename=self._temp_HWI_name
                )
            else:
                k_iter = self._calculate_invert_k(k_iter)

        if "k-size" in return_type:
            k_wrap_iter = self._convert_to_bit_size(k_iter)
        elif "hamming-weight" in return_type:
            k_wrap_iter = self._convert_to_hamming_weight(k_iter)
        else:
            raise ValueError(
                "Iterator return must be "
                "k-size[-invert] or hamming-weight[-invert]"
            )

        return k_wrap_iter

    def ecdsa_max_value(self):
        """Returns the max K size in BITS depending on the ECDSA private key"""
        return ecdsa.util.bit_length(self.priv_key.curve.order)

    def ecdh_iter(self, return_type="size"):
        """
        Iterator. Iterator to use for secret created by ECDH private keys.
        """
        secret_iter = self._get_data_from_binary_file(
            self.values,
            self.value_size if self.value_size else self.ecdh_max_value(),
            convert_to_int=True, endian=self.value_endianness)

        if "invert" in return_type and self.verbose:
            print("[w] Invert is not supported in ECDH. Skipping...")
            return None

        if "size" in return_type:
            secret_wrap_iter = self._convert_to_bit_size(secret_iter)
        elif "hamming-weight" in return_type:
            secret_wrap_iter = self._convert_to_hamming_weight(secret_iter)
        else:
            raise ValueError(
                "Iterator return must be k-size or hamming-weight"
            )

        return secret_wrap_iter

    def ecdh_max_value(self, bits=False):
        """
        Returns the max shared secret size in BYTES depending on the ECDH
        private key.
        """
        if bits:
            return ecdsa.util.bit_length(self.priv_key.curve.curve.p())
        else:
            return int(
                (ecdsa.util.bit_length(self.priv_key.curve.curve.p()) + 7) / 8)

    def _create_and_write_line(self):
        """
        Takes multiple possible values for each key value, selecting one
        in random for each key and writes the created line into the
        intermediate file.
        """
        if self._max_value not in self._current_line:
            return

        self._line_to_write = {
            self._max_value: self._current_line[self._max_value]
        }

        final_choices = {}

        # The idea here is that each comparing_value entry will choose
        # randomly if he will pair with a non comparing_value that chose it
        # from before or from after, if it has on both sides value. Finally,
        # if the side has more than one values, it choose one in random.
        for side in ['before', 'after']:
            for size in self._current_line[side]:
                if size not in final_choices:
                    final_choices[size] = {}

                num_of_values = len(self._current_line[side][size])
                if num_of_values == 1:
                    final_choices[size][side] = \
                        self._current_line[side][size][0]
                else:
                    final_choices[size][side] = \
                        choice(self._current_line[side][size])

                    self._measurements_dropped += num_of_values - 1

        for size in final_choices:
            random_choice = choice(['before', 'after'])

            if random_choice not in final_choices[size]:
                if random_choice == 'before':
                    random_choice = 'after'
                else:
                    random_choice = 'before'

            self._line_to_write[size] = final_choices[size][random_choice]
            self._selections[size][random_choice] += 1

        line = ""
        for size in self._line_to_write:
            line += '{0},{1},'.format(
                size, self._line_to_write[size]
            )
        self._intermedian_fp.write(line[:-1] + '\n')

    def _write_selections(self):
        """
        Writes how many times a value was selected, from before or after,
        into a file.
        """
        selections_keys_sorted = sorted(self._selections.keys(), reverse=True)

        with open(
            join(self.output, "selections.csv"), 'w', encoding="utf-8"
        ) as out_fp:
            out_fp.write("value,before,after\n")
            for size in selections_keys_sorted:
                out_fp.write('{0},{1},{2}\n'.format(
                    size,
                    self._selections[size]['before'],
                    self._selections[size]['after']
                ))

    def _append_sanity_to_line(self, data):
        """Appends sanity value into a measurement line."""
        self._measurements_fp.write('{0},{1},{2}\n'.format(
                                self._row, data[0], data[1]
                            ))

    def _create_and_write_sanity_entries(self):
        """
        Reads the intermediate file and adds sanity records to the final
        mesurements file.
        """
        temp_file_path = self._intermedian_fp.name
        self._intermedian_fp.close()
        state = WAIT_FOR_FIRST_BARE_MAX_VALUE
        last_single_max_value = None
        self._row -= 1
        sanity_entries_count = 0

        with open(temp_file_path, "r") as in_fp:
            reader = csv.reader(in_fp)
            for row in reader:
                if len(row) > 2:
                    self._row += 1
                    for i in range(0, len(row), 2):
                        self._measurements_fp.write(
                            '{0},{1},{2}\n'.format(
                                self._row, row[i], row[i + 1]
                            )
                        )

                    if len(row) / 2 > self._max_tuple_size:
                        self._max_tuple_size = len(row) // 2

                if state != WAIT_FOR_SECOND_BARE_MAX_VALUE and len(row) == 2:
                    if state == WAIT_FOR_NON_BARE_MAX_VALUE:
                        self._measurements_dropped += 1

                    last_single_max_value = row.copy()
                    state = WAIT_FOR_NON_BARE_MAX_VALUE
                elif state == WAIT_FOR_NON_BARE_MAX_VALUE and len(row) > 2:
                    state = WAIT_FOR_SECOND_BARE_MAX_VALUE
                elif state == WAIT_FOR_SECOND_BARE_MAX_VALUE and len(row) == 2:
                    random_choice = choice([0, 1])
                    sanity_entries_count += 1

                    if random_choice == 0: # use the one before
                        self._append_sanity_to_line(last_single_max_value)
                        last_single_max_value = row.copy()
                        state = WAIT_FOR_NON_BARE_MAX_VALUE
                    else: # use the one after
                        self._append_sanity_to_line(row)
                        state = WAIT_FOR_FIRST_BARE_MAX_VALUE
                else:
                    if state > WAIT_FOR_FIRST_BARE_MAX_VALUE:
                        self._measurements_dropped += 1

                    state = WAIT_FOR_FIRST_BARE_MAX_VALUE

        remove(temp_file_path)

        if self.verbose:
            print('[i] {0}-bit-sized sanity entries: {1:,}'.format(
                self._max_value, sanity_entries_count
            ))

    def _check_for_iter_left_overs(
            self, iterator, desc='Left-overs on iterator:'):
        left_overs_counter = 0
        for item in iterator:
            left_overs_counter += 1
            if self.verbose:
                print(item)

        if left_overs_counter > 0:
            if self.verbose:
                print(desc + " {0}".format(left_overs_counter))

            raise ValueError("There are some extra values that are not used.")

    def process_measurements_and_create_csv_file(
            self, values_iter, comparing_value
            ):
        """
        Processing all the nonces and associated time measurements from the
        given files and creates a randomized measurement file with tuples
        associating the max values with non max values.
        """
        if not all([values_iter, comparing_value]):
            return

        self._measurements_fp = open(
            join(self.output, self.measurements_csv), "w"
        )
        self._intermedian_fp = tempfile.NamedTemporaryFile(
            mode="w", delete=False
        )
        self._current_line = {
            "before": {}
        }
        self._next_line = None
        self._line_to_write = {}
        self._max_tuple_size = 0
        self._measurements_dropped = 0
        self._selections = defaultdict(lambda: defaultdict(lambda: 0))
        self._row = 0
        self._max_value = \
            self.max_bit_size if self.max_bit_size else comparing_value

        time_iter = self._get_time_from_file()

        if self.verbose:
            print("[i] Creating {0} file...".format(self.measurements_csv))

        progress = None
        status = [0]
        if self.verbose and self._total_measurements:
            status = [0, self._total_measurements, Event()]
            kwargs = {}
            kwargs['unit'] = ' signatures'
            kwargs['prefix'] = 'decimal'
            kwargs['delay'] = self.delay
            kwargs['end'] = self.carriage_return
            progress = Thread(target=progress_report, args=(status,),
                            kwargs=kwargs)
            progress.start()

        try:
            for value, time_value in izip(values_iter, time_iter):
                status[0] += 1

                # The idea here is that every value != comparing_value chooses
                # randomly to pair with the comparing_value before or with
                # the one after
                if value == comparing_value:
                    if self._next_line is None:
                        self._current_line[value] = time_value
                        self._current_line['after'] = {}
                        self._next_line = {
                            'before': {}
                        }
                    else:
                        self._create_and_write_line()

                        self._current_line = self._next_line
                        self._current_line[value] = time_value
                        self._current_line['after'] = {}
                        self._next_line = {
                            'before': {}
                        }
                else:
                    if self._next_line is not None:
                        random_choice = choice([0, 1])
                        if random_choice == 0:
                            line = self._current_line
                        else:
                            line = self._next_line
                    else:
                        line = self._current_line

                    if "after" in line:
                        stage = "after"
                    else:
                        stage = "before"

                    if value not in line[stage]:
                        line[stage][value] = []

                    line[stage][value].append(time_value)

            if self._next_line:
                for size in self._next_line["before"]:
                    self._measurements_dropped += \
                        len(self._next_line["before"][size])

            self._create_and_write_line()

            self._write_selections()
        finally:
            if progress:
                status[2].set()
                progress.join()
                print()

        self._check_for_iter_left_overs(
            values_iter, "Left over values in measurements"
        )

        self._check_for_iter_left_overs(
            time_iter, "Left over times in measurements"
        )

        self._create_and_write_sanity_entries()

        if self.verbose:
            if self._total_measurements:
                print(
                    '[i] Measurements that have been dropped: {0:,} ({1:.2f}%)'
                    .format(
                        self._measurements_dropped,
                        (self._measurements_dropped * 100)
                        / self._total_measurements
                    )
                )
            print(
                '[i] Biggest tuple size in file: {0}\n'
                    .format(self._max_tuple_size) +
                '[i] Written rows: {0:,}'.format(max(0, self._row))
            )

        self._measurements_fp.close()

    def _write_hamming_weight_line(self, row, line_to_write):
        """
        Takes one or more possible Hamming weight values for each key value,
        selecting one in random for each key and writes the created line into
        the measurements file.
        """
        measurements_dropped = 0

        for key in line_to_write:
            measurements_dropped += len(line_to_write[key]) - 1
            line_to_write[key] = choice(line_to_write[key])
            self._measurements_fp.write("{0},{1},{2}\n".format(
                row, key, line_to_write[key]))

        return measurements_dropped

    def process_measurements_and_create_hamming_csv_file(
            self, values_iter, items_in_tuple = 20):
        """
        Processing all the nonces and associated time measurements from the
        given files and creates a file with tuples associating the Hamming
        weight of the nonces.
        """
        if not values_iter:
            return

        self._measurements_fp = open(
            join(self.output, self.measurements_csv), "w"
        )

        line_to_write = defaultdict(list)
        items_buffered = 0
        row = 0
        measurements_dropped = 0
        min_tuple_size = items_in_tuple
        last_tuple_size = 0

        time_iter = self._get_time_from_file()

        if self.verbose:
            print("[i] Creating {0} file...".format(self.measurements_csv))

        progress = None
        status = [0]
        if self.verbose and self._total_measurements:
            status = [0, self._total_measurements, Event()]
            kwargs = {}
            kwargs['unit'] = ' signatures'
            kwargs['prefix'] = 'decimal'
            kwargs['delay'] = self.delay
            kwargs['end'] = self.carriage_return
            progress = Thread(target=progress_report, args=(status,),
                            kwargs=kwargs)
            progress.start()

        try:
            for value, time_value in izip(values_iter, time_iter):
                status[0] += 1

                line_to_write[value].append(time_value)
                items_buffered += 1

                if items_buffered == items_in_tuple:
                    line_measurements_drop = self._write_hamming_weight_line(
                        row, line_to_write
                    )

                    min_tuple_size = min(
                        len(line_to_write.keys()), min_tuple_size)
                    measurements_dropped += line_measurements_drop
                    row += 1
                    line_to_write = defaultdict(list)
                    items_buffered = 0
        finally:
            if progress:
                status[2].set()
                progress.join()
                print()

        if items_buffered > 0:
            line_measurements_drop = self._write_hamming_weight_line(
                row, line_to_write
            )

            last_tuple_size = len(line_to_write.keys())
            measurements_dropped += line_measurements_drop
            row += 1

        self._check_for_iter_left_overs(values_iter)
        self._check_for_iter_left_overs(time_iter)

        if self.verbose:
            if self._total_measurements:
                print(
                    '[i] Measurements that have been dropped: {0:,} ({1:.2f}%)'
                    .format(
                        measurements_dropped,
                        (measurements_dropped * 100)
                        / self._total_measurements
                    )
                )

            if last_tuple_size > 0:
                print(
                    '[i] Smallest non-last tuple size in file: {0}\n'
                        .format(min_tuple_size) +
                    '[i] Last tuple size in file: {0}'.format(last_tuple_size)
                )
            else:
                print(
                    '[i] Smallest tuple size in file: {0}'
                        .format(min_tuple_size)
                )

            print('[i] Written rows: {0:,}'.format(row))

        self._measurements_fp.close()

    def process_and_create_multiple_csv_files(self, files = {
        "measurements.csv": "k-size",
    }, ecdh = False):
        original_measuremments_csv = self.measurements_csv
        skipped_h_weight_invert = False
        h_weight_invert_file = None
        h_weight_invert_mode = None

        if ecdh:
            self._total_measurements = int(
                getsize(self.data) / ((2 * self.ecdh_max_value()) + 1))
            for file in files:
                self.measurements_csv = file

                if "hamming-weight" in files[file]:
                    self.process_measurements_and_create_hamming_csv_file(
                        self.ecdh_iter(return_type=files[file])
                    )
                else:
                    self.process_measurements_and_create_csv_file(
                        self.ecdh_iter(return_type=files[file]),
                        self.ecdh_max_value(bits=True)
                    )
            return

        if exists(join(self.output, "ecdsa-k-time-map.csv")):
            remove(join(self.output, "ecdsa-k-time-map.csv"))

        for file, mode in files.items():
            if "hamming-weight" in mode and "invert" in mode:
                skipped_h_weight_invert = True
                h_weight_invert_file = file
                h_weight_invert_mode = mode
                self._temp_HWI_name = join(self.output, "tmp_HWI_values.csv")

        for file in files:
            self.measurements_csv = file

            if "hamming-weight" in files[file]:
                if "invert" in files[file]:
                    continue

                self.process_measurements_and_create_hamming_csv_file(
                    self.ecdsa_iter(return_type=files[file])
                )
            else:
                self.process_measurements_and_create_csv_file(
                    self.ecdsa_iter(return_type=files[file]),
                    self.ecdsa_max_value()
                )

        if skipped_h_weight_invert:
            self.measurements_csv = h_weight_invert_file
            self.process_measurements_and_create_hamming_csv_file(
                self.ecdsa_iter(return_type=h_weight_invert_mode)
            )
            remove(self._temp_HWI_name)
            self._temp_HWI_name = None

        self.measurements_csv = original_measuremments_csv

    @staticmethod
    def hostname_to_ip(hostname):
        """
        Converts hostname to IPv4 address, if needed.
        :param str hostname: hostname or an IPv4 address
        :return: str IPv4 address
        """
        # first check if it is not already IPv4
        try:
            ip = inet_aton(hostname)
            return ip
        except error:
            pass

        # not an IPv4, try a hostname
        try:
            ip = gethostbyname(hostname)
            return inet_aton(ip)
        except gaierror:
            raise Exception("Hostname is not an IPv4 or a reachable hostname")

    def _read_private_key(self, file):
        lines = []
        while True:
            line = file.readline()
            # empty line still has '\n', only EOF is an empty string
            if not line:
                return None
            line = line.strip()
            if line == "-----BEGIN PRIVATE KEY-----":
                lines.append(line)
                break
        while True:
            line = file.readline()
            if not line:
                raise ValueError("Truncated private key file!")
            line = line.strip()
            if line == "-----BEGIN PRIVATE KEY-----":
                raise ValueError("Inconsistent private key file!")
            lines.append(line)
            if line == "-----END PRIVATE KEY-----":
                break

        one_pem_key = "\n".join(lines)

        return Python_Key.parsePEM(one_pem_key)

    def process_rsa_keys(self):
        # list of values for the Hamming weight of d, p, q, dP, dQ, qInv
        values = []
        times = []
        max_len = 20

        tuple_num = 0

        value_names = ('d', 'p', 'q', 'dP', 'dQ', 'qInv')

        rsa_keys = None
        measurements = dict((i, None) for i in value_names)

        times_iterator = self._get_time_from_file()

        try:
            rsa_keys = open(self.rsa_keys, "rt")
            for i in value_names:
                f_name = join(self.output, 'measurements-' + i + '.csv')
                measurements[i] = open(f_name, 'wt')

            while True:
                # read an RSA private key
                key = self._read_private_key(rsa_keys)
                if key:
                    # extract Hamming weights of the private key parameters
                    values.append(dict((i, bit_count(getattr(key, i)))
                                       for i in
                                       value_names))
                    times.append(next(times_iterator))

                # once we have few measurements collect them into tuples
                # and write to files
                if len(values) >= max_len or (not key and times):
                    for v_n in value_names:
                        keys = set(v[v_n] for v in values)
                        size_and_time = sorted(zip(
                            (v[v_n] for v in values), times))

                        for k in sorted(keys):
                            to_select = [i for i in size_and_time if i[0] == k]
                            # since sometimes for the same key we can have
                            # multiple values, write a randomly selected one
                            selected = choice(to_select)
                            measurements[v_n].write("{0},{1},{2}\n".format(
                                tuple_num, selected[0], selected[1]))

                    values = []
                    times = []
                    tuple_num += 1

                if not key:
                    break

        finally:
            if rsa_keys:
                rsa_keys.close()
            for i in value_names:
                if measurements[i]:
                    measurements[i].close()

if __name__ == '__main__':
    main()
